from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union, Optional

from wrapt import ObjectProxy  # type: ignore

from payi.lib.helpers import _compact_json, _set_attr_safe
from payi.types.shared.xproxy_error import XproxyError
from payi.types.shared.xproxy_result import XproxyResult

from .ProviderRequest import _ChunkResult, _ProviderRequest

if TYPE_CHECKING:
    from .instrument import _PayiInstrumentor

__all__ = [
    "_StreamIteratorWrapper",
    "_StreamManagerWrapper",
    "_GeneratorWrapper",
]

class _StreamIteratorWrapper(ObjectProxy):  # type: ignore
    def __init__(
        self,
        response: Any,
        instance: Any,
        instrumentor: '_PayiInstrumentor',
        request: _ProviderRequest,
    ) -> None:

        instrumentor._logger.debug(f"StreamIteratorWrapper: instance {instance}, category {request._category}")

        request.process_initial_stream_response(response)

        bedrock_from_stream: bool = False
        if request.is_aws_client:
            stream = response.get("stream", None)

            if stream:
                response = stream
                bedrock_from_stream = True
            else:
                response = response.get("body")
                bedrock_from_stream = False

        super().__init__(response)  # type: ignore

        self._response = response
        self._instance = instance

        self._instrumentor = instrumentor
        self._responses: list[str] = []

        self._request: _ProviderRequest = request

        self._first_token: bool = True
        self._bedrock_from_stream: bool = bedrock_from_stream
        self._ingested: bool = False
        self._iter_started: bool = False
        self._log_prompt_and_response: bool = request._log_prompt_and_response

    def __enter__(self) -> Any:
        self._instrumentor._logger.debug(f"StreamIteratorWrapper: __enter__")
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None: 
        self._instrumentor._logger.debug(f"StreamIteratorWrapper: __exit__")
        self.__wrapped__.__exit__(exc_type, exc_val, exc_tb)  # type: ignore

    async def __aenter__(self) -> Any:
        self._instrumentor._logger.debug(f"StreamIteratorWrapper: __aenter__")
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self._instrumentor._logger.debug(f"StreamIteratorWrapper: __aexit__")
        await self.__wrapped__.__aexit__(exc_type, exc_val, exc_tb)  # type: ignore

    def __iter__(self) -> Any:  
        self._iter_started = True
        if self._request.is_aws_client:
            # MUST reside in a separate function so that the yield statement (e.g. the generator) doesn't implicitly return its own iterator and overriding self
            self._instrumentor._logger.debug(f"StreamIteratorWrapper: bedrock __iter__")
            return self._iter_bedrock()

        self._instrumentor._logger.debug(f"StreamIteratorWrapper: __iter__")
        return self

    def _iter_bedrock(self) -> Any:
        # botocore EventStream doesn't have a __next__ method so iterate over the wrapped object in place
        for event in self.__wrapped__: # type: ignore
            result: Optional[_ChunkResult] = None

            if (self._bedrock_from_stream):
                result = self._evaluate_chunk(event)
            else:
                chunk = event.get('chunk') # type: ignore
                if chunk:
                    decode = chunk.get('bytes').decode() # type: ignore
                    result = self._evaluate_chunk(decode)

            if result and result.ingest:
                from .BedrockInstrumentor import BedrockInstrumentor

                xproxy_result = self._stop_iteration()

                # the xproxy_result is not json serializable by default so adding the object is opt in by the client
                if BedrockInstrumentor._add_streaming_xproxy_result:
                    self._request.assign_xproxy_result(event, xproxy_result)
            yield event

        self._instrumentor._logger.debug(f"StreamIteratorWrapper: bedrock iter finished")

        self._stop_iteration()

    def __aiter__(self) -> Any:
        self._iter_started = True
        self._instrumentor._logger.debug(f"StreamIteratorWrapper: __aiter__")
        return self

    def __next__(self) -> object:
        try:
            chunk: object = self.__wrapped__.__next__()  # type: ignore

            if self._ingested:
                self._instrumentor._logger.debug(f"StreamIteratorWrapper: __next__ already ingested, not processing chunk {chunk}")
                return chunk # type: ignore

            result = self._evaluate_chunk(chunk)

            if result.ingest:
                xproxy_result = self._stop_iteration()
                self._request.assign_xproxy_result(chunk, xproxy_result)

            if result.send_chunk_to_caller:
                return chunk # type: ignore
            else:
                return self.__next__()
        except Exception as e:
            if isinstance(e, StopIteration):
                self._stop_iteration()
            else:
                self._instrumentor._logger.debug(f"StreamIteratorWrapper: __next__ exception {e}")
            raise e

    async def __anext__(self) -> object:
        try:
            chunk: object = await self.__wrapped__.__anext__()  # type: ignore

            if self._ingested:
                self._instrumentor._logger.debug(f"StreamIteratorWrapper: __next__ already ingested, not processing chunk {chunk}")
                return chunk # type: ignore

            result = self._evaluate_chunk(chunk)

            if result.ingest:
                xproxy_result = await self._astop_iteration()
                self._request.assign_xproxy_result(chunk, xproxy_result)

            if  result.send_chunk_to_caller:
                return chunk # type: ignore
            else:
                return await self.__anext__()

        except Exception as e:
            if isinstance(e, StopAsyncIteration):
                await self._astop_iteration()
            else:
                self._instrumentor._logger.debug(f"StreamIteratorWrapper: __anext__ exception {e}")
            raise e

    def _evaluate_chunk(self, chunk: Any) -> _ChunkResult:
        if self._first_token:
            self._request._ingest["time_to_first_token_ms"] = self._request.stopwatch.elapsed_ms_int()
            self._first_token = False

        if self._log_prompt_and_response:
            self._responses.append(self.chunk_to_json(chunk))

        return self._request.process_chunk(chunk)

    def _process_stop_iteration(self) -> None:
        self._instrumentor._logger.debug(f"StreamIteratorWrapper: process stop iteration")

        self._request.stopwatch.stop()
        self._request._ingest["end_to_end_latency_ms"] = self._request.stopwatch.elapsed_ms_int()
        self._request._ingest["http_status_code"] = 200

        if self._log_prompt_and_response:
            self._request._ingest["provider_response_json"] = self._responses

    async def _astop_iteration(self) -> Optional[Union[XproxyResult, XproxyError]]:
        if self._ingested:
            self._instrumentor._logger.debug(f"StreamIteratorWrapper: astop iteration already ingested, skipping")
            return None

        self._process_stop_iteration()
        xproxy_result = await self._instrumentor._aingest_units(self._request)
        self._ingested = True

        return xproxy_result

    def _stop_iteration(self) -> Optional[Union[XproxyResult, XproxyError]]:
        if self._ingested:
            self._instrumentor._logger.debug(f"StreamIteratorWrapper: stop iteration already ingested, skipping")
            return None

        self._process_stop_iteration()
        xproxy_result = self._instrumentor._ingest_units(self._request)
        self._ingested = True

        return xproxy_result

    @staticmethod
    def chunk_to_json(chunk: Any) -> str:
        if hasattr(chunk, "to_json"):
            return str(chunk.to_json())
        elif isinstance(chunk, bytes):
            return chunk.decode()
        elif isinstance(chunk, str):
            return chunk
        else:
            # assume dict
            return _compact_json(chunk)

class _StreamManagerWrapper(ObjectProxy):  # type: ignore
    def __init__(
        self,
        stream_manager: Any,  # type: ignore
        instance: Any,
        instrumentor: _PayiInstrumentor, 
        request: _ProviderRequest,
    ) -> None:
        instrumentor._logger.debug(f"StreamManagerWrapper: instance {instance}, category {request._category}")

        super().__init__(stream_manager)  # type: ignore

        self._stream_manager = stream_manager  
        self._instance = instance
        self._instrumentor = instrumentor
        self._responses: list[str] = []
        self._request: _ProviderRequest = request
        self._first_token: bool = True

    def __enter__(self) -> Any:
        self._instrumentor._logger.debug(f"_StreamManagerWrapper: __enter__")

        # Underlying iterator is wrapped separately, expects attr _payi_request to be set
        stream = self.__wrapped__.__enter__()  # type: ignore
        
        # Attach tracking info
        _set_attr_safe(stream, '_payi_request', self._request)
        return stream # type: ignore

    async def __aenter__(self) -> Any:
        self._instrumentor._logger.debug(f"_StreamManagerWrapper: __aenter__")

        stream = await self.__wrapped__.__aenter__()  # type: ignore
        
        # Attach tracking info
        _set_attr_safe(stream, '_payi_request', self._request)
        return stream # type: ignore

class _GeneratorWrapper:  # type: ignore
    def __init__(
        self,
        generator: Any,
        instance: Any,
        instrumentor: _PayiInstrumentor, 
        request: _ProviderRequest,
    ) -> None:
        instrumentor._logger.debug(f"GeneratorWrapper: instance {instance}, category {request._category}")

        super().__init__()  # type: ignore
        
        self._generator = generator
        self._instance = instance
        self._instrumentor = instrumentor
        self._log_prompt_and_response: bool = request._log_prompt_and_response
        self._responses: list[str] = []
        self._request: _ProviderRequest = request
        self._first_token: bool = True
        self._ingested: bool = False
        self._iter_started: bool = False

    def __iter__(self) -> Any:
        self._iter_started = True
        self._instrumentor._logger.debug(f"GeneratorWrapper: __iter__")
        return self
        
    def __aiter__(self) -> Any:
        self._instrumentor._logger.debug(f"GeneratorWrapper: __aiter__")
        return self

    def _process_chunk(self, chunk: Any) -> _ChunkResult:
        if self._first_token:
            self._request._ingest["time_to_first_token_ms"] = self._request.stopwatch.elapsed_ms_int()
            self._first_token = False
            
        if self._log_prompt_and_response:
            dict = self._chunk_to_dict(chunk) 
            self._responses.append(_compact_json(dict))
                
        return self._request.process_chunk(chunk)
    
    def __next__(self) -> Any:
        try:
            chunk = next(self._generator)
            result = self._process_chunk(chunk)

            if result.ingest:
                xproxy_result = self._stop_iteration()
                self._request.assign_xproxy_result(chunk, xproxy_result)

            # ignore result.send_chunk_to_caller:
            return chunk

        except Exception as e:
            if isinstance(e, StopIteration):
                self._stop_iteration()
            else:
                self._instrumentor._logger.debug(f"GeneratorWrapper: __next__ exception {e}")            
            raise e

    async def __anext__(self) -> Any:
        try:
            chunk = await anext(self._generator) # type: ignore
            result = self._process_chunk(chunk)

            if result.ingest:
                xproxy_result = await self._astop_iteration()
                self._request.assign_xproxy_result(chunk, xproxy_result)

            # ignore result.send_chunk_to_caller:
            return chunk # type: ignore

        except Exception as e:
            if isinstance(e, StopAsyncIteration):
                await self._astop_iteration()
            else:
                self._instrumentor._logger.debug(f"GeneratorWrapper: __anext__ exception {e}")
            raise e

    @staticmethod
    def _chunk_to_dict(chunk: Any) -> 'dict[str, object]':
        if hasattr(chunk, "to_dict"):
            return chunk.to_dict() # type: ignore
        elif hasattr(chunk, "to_json_dict"):  
            return chunk.to_json_dict() # type: ignore
        else:
            return {}

    def _stop_iteration(self) -> Optional[Union[XproxyResult, XproxyError]]:
        if self._ingested:
            self._instrumentor._logger.debug(f"GeneratorWrapper: stop iteration already ingested, skipping")
            return None

        self._process_stop_iteration()
        xproxy_result = self._instrumentor._ingest_units(self._request)
        self._ingested = True
        return xproxy_result

    async def _astop_iteration(self) -> Optional[Union[XproxyResult, XproxyError]]:
        if self._ingested:
            self._instrumentor._logger.debug(f"GeneratorWrapper: astop iteration already ingested, skipping")
            return None

        self._process_stop_iteration()
        xproxy_result = await self._instrumentor._aingest_units(self._request)
        self._ingested = True
        return xproxy_result

    def _process_stop_iteration(self) -> None:
        self._instrumentor._logger.debug(f"GeneratorWrapper: stop iteration")

        self._request.stopwatch.stop()
        self._request._ingest["end_to_end_latency_ms"] = self._request.stopwatch.elapsed_ms_int()
        self._request._ingest["http_status_code"] = 200
            
        if self._log_prompt_and_response:
            self._request._ingest["provider_response_json"] = self._responses




# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Optional
from datetime import datetime
from typing_extensions import Literal

from pydantic import Field as FieldInfo

from .._models import BaseModel

__all__ = [
    "CategoryResourceResponse",
    "Units",
    "AwsBedrockResource",
    "AzureResource",
    "GoogleVertexResource",
    "MappedResource",
]


class Units(BaseModel):
    input_price: Optional[float] = None

    output_price: Optional[float] = None


class AwsBedrockResource(BaseModel):
    aws_model_units: int = FieldInfo(alias="model_units")


class AzureResource(BaseModel):
    ptus: int


class GoogleVertexResource(BaseModel):
    gsus: int


class MappedResource(BaseModel):
    category: Optional[str] = None

    resource: Optional[str] = None

    scope: Optional[Literal["global", "datazone", "region"]] = None

    sub_scope: Optional[str] = None


class CategoryResourceResponse(BaseModel):
    active: bool

    category: str

    proxy_allowed: bool

    resource: str

    resource_id: str

    start_timestamp: datetime

    units: Dict[str, Units]

    aliased_resource: Optional[str] = None

    aws_bedrock_resource: Optional[AwsBedrockResource] = None

    azure_resource: Optional[AzureResource] = None

    character_billing: Optional[bool] = None

    cost_per_hour: Optional[float] = None

    deprecated_timestamp: Optional[datetime] = None

    description: Optional[str] = None

    end_timestamp: Optional[datetime] = None

    google_vertex_resource: Optional[GoogleVertexResource] = None

    large_context_threshold: Optional[int] = None

    mapped_resource: Optional[MappedResource] = None

    max_input_units: Optional[int] = None

    max_output_units: Optional[int] = None

    max_total_units: Optional[int] = None

    reservation_id: Optional[str] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import httpx

from ...._types import Body, Query, Headers, NotGiven, not_given
from ...._compat import cached_property
from ...._resource import SyncAPIResource, AsyncAPIResource
from ...._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ...._base_client import make_request_options
from ....types.use_cases.use_case_definition_response import UseCaseDefinitionResponse

__all__ = ["VersionResource", "AsyncVersionResource"]


class VersionResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> VersionResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return VersionResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> VersionResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return VersionResourceWithStreamingResponse(self)

    def increment(
        self,
        use_case_name: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> UseCaseDefinitionResponse:
        """
        Increment a Use Case version

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not use_case_name:
            raise ValueError(f"Expected a non-empty value for `use_case_name` but received {use_case_name!r}")
        return self._post(
            f"/api/v1/use_cases/definitions/{use_case_name}/increment_version",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=UseCaseDefinitionResponse,
        )


class AsyncVersionResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncVersionResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return AsyncVersionResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncVersionResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return AsyncVersionResourceWithStreamingResponse(self)

    async def increment(
        self,
        use_case_name: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> UseCaseDefinitionResponse:
        """
        Increment a Use Case version

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not use_case_name:
            raise ValueError(f"Expected a non-empty value for `use_case_name` but received {use_case_name!r}")
        return await self._post(
            f"/api/v1/use_cases/definitions/{use_case_name}/increment_version",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=UseCaseDefinitionResponse,
        )


class VersionResourceWithRawResponse:
    def __init__(self, version: VersionResource) -> None:
        self._version = version

        self.increment = to_raw_response_wrapper(
            version.increment,
        )


class AsyncVersionResourceWithRawResponse:
    def __init__(self, version: AsyncVersionResource) -> None:
        self._version = version

        self.increment = async_to_raw_response_wrapper(
            version.increment,
        )


class VersionResourceWithStreamingResponse:
    def __init__(self, version: VersionResource) -> None:
        self._version = version

        self.increment = to_streamed_response_wrapper(
            version.increment,
        )


class AsyncVersionResourceWithStreamingResponse:
    def __init__(self, version: AsyncVersionResource) -> None:
        self._version = version

        self.increment = async_to_streamed_response_wrapper(
            version.increment,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from payi import Payi, AsyncPayi
from payi.types import UseCaseInstanceResponse
from tests.utils import assert_matches_type

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestProperties:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_update(self, client: Payi) -> None:
        property = client.use_cases.properties.update(
            use_case_id="use_case_id",
            use_case_name="use_case_name",
            properties={"foo": "string"},
        )
        assert_matches_type(UseCaseInstanceResponse, property, path=["response"])

    @parametrize
    def test_raw_response_update(self, client: Payi) -> None:
        response = client.use_cases.properties.with_raw_response.update(
            use_case_id="use_case_id",
            use_case_name="use_case_name",
            properties={"foo": "string"},
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        property = response.parse()
        assert_matches_type(UseCaseInstanceResponse, property, path=["response"])

    @parametrize
    def test_streaming_response_update(self, client: Payi) -> None:
        with client.use_cases.properties.with_streaming_response.update(
            use_case_id="use_case_id",
            use_case_name="use_case_name",
            properties={"foo": "string"},
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            property = response.parse()
            assert_matches_type(UseCaseInstanceResponse, property, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_update(self, client: Payi) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `use_case_name` but received ''"):
            client.use_cases.properties.with_raw_response.update(
                use_case_id="use_case_id",
                use_case_name="",
                properties={"foo": "string"},
            )

        with pytest.raises(ValueError, match=r"Expected a non-empty value for `use_case_id` but received ''"):
            client.use_cases.properties.with_raw_response.update(
                use_case_id="",
                use_case_name="use_case_name",
                properties={"foo": "string"},
            )


class TestAsyncProperties:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_update(self, async_client: AsyncPayi) -> None:
        property = await async_client.use_cases.properties.update(
            use_case_id="use_case_id",
            use_case_name="use_case_name",
            properties={"foo": "string"},
        )
        assert_matches_type(UseCaseInstanceResponse, property, path=["response"])

    @parametrize
    async def test_raw_response_update(self, async_client: AsyncPayi) -> None:
        response = await async_client.use_cases.properties.with_raw_response.update(
            use_case_id="use_case_id",
            use_case_name="use_case_name",
            properties={"foo": "string"},
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        property = await response.parse()
        assert_matches_type(UseCaseInstanceResponse, property, path=["response"])

    @parametrize
    async def test_streaming_response_update(self, async_client: AsyncPayi) -> None:
        async with async_client.use_cases.properties.with_streaming_response.update(
            use_case_id="use_case_id",
            use_case_name="use_case_name",
            properties={"foo": "string"},
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            property = await response.parse()
            assert_matches_type(UseCaseInstanceResponse, property, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_update(self, async_client: AsyncPayi) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `use_case_name` but received ''"):
            await async_client.use_cases.properties.with_raw_response.update(
                use_case_id="use_case_id",
                use_case_name="",
                properties={"foo": "string"},
            )

        with pytest.raises(ValueError, match=r"Expected a non-empty value for `use_case_id` but received ''"):
            await async_client.use_cases.properties.with_raw_response.update(
                use_case_id="",
                use_case_name="use_case_name",
                properties={"foo": "string"},
            )



from __future__ import annotations

from typing import Any, List, Union, Sequence
from typing_extensions import override

from wrapt import wrap_function_wrapper  # type: ignore

from .instrument import _IsStreaming, _PayiInstrumentor
from .VertexRequest import _VertexRequest
from .version_helper import get_version_helper
from .ProviderRequest import _ChunkResult


class GoogleGenAiInstrumentor:
    _module_name: str = "google-genai"
    _module_version: str = ""

    @staticmethod
    def instrument(instrumentor: _PayiInstrumentor) -> None:
        GoogleGenAiInstrumentor._module_version = get_version_helper(GoogleGenAiInstrumentor._module_name)

        wrappers = [
            ("google.genai.models", "Models.generate_content", generate_wrapper(instrumentor)),
            ("google.genai.models", "Models.generate_content_stream", generate_stream_wrapper(instrumentor)),
            ("google.genai.models", "AsyncModels.generate_content", agenerate_wrapper(instrumentor)),
            ("google.genai.models", "AsyncModels.generate_content_stream", agenerate_stream_wrapper(instrumentor)),
        ]

        for module, method, wrapper in wrappers:
            try:
                wrap_function_wrapper(module, method, wrapper)
            except Exception as e:
                instrumentor._logger.debug(f"Failed to wrap {module}.{method}: {e}")

@_PayiInstrumentor.payi_wrapper
def generate_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("genai generate_content wrapper")
    return instrumentor.invoke_wrapper(
        _GoogleGenAiRequest(instrumentor=instrumentor, instance=instance),
        _IsStreaming.false,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_wrapper
def generate_stream_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("genai generate_content_stream wrapper")
    return instrumentor.invoke_wrapper(
        _GoogleGenAiRequest(instrumentor=instrumentor, instance=instance),
        _IsStreaming.true,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_awrapper
async def agenerate_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("async genai generate_content wrapper")
    return await instrumentor.async_invoke_wrapper(
        _GoogleGenAiRequest(instrumentor=instrumentor, instance=instance),
        _IsStreaming.false,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_wrapper
async def agenerate_stream_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("async genai generate_content_stream wrapper")
    return await instrumentor.async_invoke_wrapper(
        _GoogleGenAiRequest(instrumentor=instrumentor, instance=instance),
        _IsStreaming.true,
        wrapped,
        instance,
        args,
        kwargs,
    )

class _GoogleGenAiRequest(_VertexRequest):
    def __init__(self, instrumentor: _PayiInstrumentor, instance: Any):
        super().__init__(
            instrumentor=instrumentor,
            instance=instance,
            module_name=GoogleGenAiInstrumentor._module_name,
            module_version=GoogleGenAiInstrumentor._module_version,
            )
        self._prompt_character_count = 0
        self._candidates_character_count = 0

    @override
    def process_request(self, instance: Any, args: Sequence[Any], kwargs: Any) -> bool:
        from google.genai.types import Content, PIL_Image, Part # type: ignore #  noqa: F401  I001

        if not kwargs:
            return True

        model: str = kwargs.get("model", "")
        self._ingest["resource"] = "google." + model

        value: Union[ # type: ignore
            Content,
            str,
            PIL_Image,
            Part,
            List[Union[str, PIL_Image, Part]],
        ] = kwargs.get("contents", None)  # type: ignore 

        items: List[Union[str, Image, Part]] = [] # type: ignore #  noqa: F401  I001

        if not value:
            raise TypeError("value must not be empty")

        if isinstance(value, Content):
            items = value.parts # type: ignore
        if isinstance(value, (str, PIL_Image, Part)):
            items = [value] # type: ignore
        if isinstance(value, list):
            items = value # type: ignore

        for item in items: # type: ignore
            text = ""
            if isinstance(item, Part):
                d = item.to_json_dict() # type: ignore
                if "text" in d:
                    text = d["text"] # type: ignore
            elif isinstance(item, str):
                text = item

            if text != "":
                self._prompt_character_count += self.count_chars_skip_spaces(text) # type: ignore

        return True

    @override
    def process_request_prompt(self, prompt: 'dict[str, Any]', args: Sequence[Any], kwargs: 'dict[str, Any]') -> None:
        from google.genai.types import Content, PIL_Image, Part, Tool, GenerateContentConfig, GenerateContentConfigDict, ToolConfig  # type: ignore #  noqa: F401  I001

        key = "contents"

        if not kwargs:
            return
        
        value: Union[ # type: ignore
            Content,
            str,
            PIL_Image,
            Part,
            List[Union[str, PIL_Image, Part]],
        ] = kwargs.get("contents", None)  # type: ignore

        items: List[Union[str, Image, Part]] = [] # type: ignore #  noqa: F401  I001

        if not value:
            return

        if isinstance(value, str):
            prompt[key] = Content(parts=[Part.from_text(text=value)]).to_json_dict() # type: ignore
        elif isinstance(value, (PIL_Image, Part)):
            prompt[key] = Content(parts=[value]).to_json_dict() # type: ignore
        elif isinstance(value, Content):
            prompt[key] = value.to_json_dict() # type: ignore
        elif isinstance(value, list):
            items = value # type: ignore
            parts = []

            for item in items: # type: ignore
                if isinstance(item, str):
                    parts.append(Part.from_text(text=item)) # type: ignore
                elif isinstance(item, Part):
                    parts.append(item) # type: ignore
                # elif isinstance(item, PIL_Image): TODO
                #     parts.append(Part.from_image(item)) # type: ignore
                        
            prompt[key] = Content(parts=parts).to_json_dict() # type: ignore

        # tools: Optional[list[Tool]] = kwargs.get("tools", None)  # type: ignore
        # if tools:
        #     t: list[dict[Any, Any]] = []
        #     for tool in tools: # type: ignore
        #         if isinstance(tool, Tool):
        #             t.append(tool.text=())  # type: ignore
        #     if t:
        #         prompt["tools"] = t
        config_kwarg = kwargs.get("config", None)  # type: ignore
        if config_kwarg is None:
            return
        
        config: GenerateContentConfigDict = {}
        if isinstance(config_kwarg, GenerateContentConfig):
            config = config_kwarg.to_json_dict()  # type: ignore
        else:
            config = config_kwarg
        
        tools = config.get("tools", None)  # type: ignore
        if isinstance(tools, list):
            t: list[dict[str, object]] = []
            for tool in tools: # type: ignore
                if isinstance(tool, Tool):
                    t.append(tool.to_json_dict())  # type: ignore
            if t:
                prompt["tools"] = t

        tool_config = config.get("tool_config", None)  # type: ignore
        if isinstance(tool_config, ToolConfig):
            prompt["tool_config"] = tool_config.to_json_dict()  # type: ignore
        elif isinstance(tool_config, dict):
            prompt["tool_config"] = tool_config

    @override
    def process_chunk(self, chunk: Any) -> _ChunkResult:
        response_dict: dict[str, Any] = chunk.to_json_dict()

        model: str = response_dict.get("model_version", "")
        self._ingest["resource"] = "google." + model

        return self.process_chunk_dict(response_dict=response_dict)

    @override
    def process_synchronous_response(
        self,
        response: Any,
        kwargs: Any) -> Any:

        return self.vertex_process_synchronous_response(response_dict=response.to_json_dict())


# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, Annotated, TypedDict

from ..._utils import PropertyInfo

__all__ = ["ResourceListParams"]


class ResourceListParams(TypedDict, total=False):
    category: Required[str]

    active: Annotated[bool, PropertyInfo(alias="Active")]

    cursor: str

    limit: int

    sort_ascending: bool



from __future__ import annotations

import json
from typing import Any, Dict, Union, Optional, Sequence
from typing_extensions import override

import tiktoken
from wrapt import wrap_function_wrapper  # type: ignore

from payi.lib.helpers import PayiCategories, PayiResourceScopes
from payi.types.ingest_units_params import IngestUnits

from .instrument import PayiInstrumentAnthropicConfig, _Context, _IsStreaming, _PayiInstrumentor
from .StreamWrappers import _GeneratorWrapper
from .version_helper import get_version_helper
from .ProviderRequest import _ChunkResult, _StreamingType, _ProviderRequest


class AnthropicInstrumentor:
    _module_name: str = "anthropic"
    _module_version: str = ""

    _azure_deployments: Dict[str, _Context] = {}
    _azure_foundry_clients_supported: bool = True

    @staticmethod
    def is_vertex(anthropic_client: Any) -> bool:
        from anthropic import AnthropicVertex, AsyncAnthropicVertex  # type: ignore # noqa: I001

        return isinstance(anthropic_client, (AsyncAnthropicVertex, AnthropicVertex))

    @staticmethod
    def is_bedrock(anthropic_client: Any) -> bool:
        from anthropic import AnthropicBedrock, AsyncAnthropicBedrock  # type: ignore # noqa: I001

        return isinstance(anthropic_client, (AsyncAnthropicBedrock, AnthropicBedrock))

    @staticmethod
    def is_azure(anthropic_client: Any) -> bool:
        if not AnthropicInstrumentor._azure_foundry_clients_supported:
            return False

        try:
            from anthropic import AnthropicFoundry, AsyncAnthropicFoundry  # type: ignore # noqa: I001
            return isinstance(anthropic_client, (AsyncAnthropicFoundry, AnthropicFoundry))
        except Exception:
            AnthropicInstrumentor._azure_foundry_clients_supported = False
            return False

    @staticmethod
    def configure(anthropic_config: PayiInstrumentAnthropicConfig) -> None:
        azure_config = anthropic_config.get("azure", {})
        if azure_config:
            model_mappings = azure_config.get("model_mappings", [])
            AnthropicInstrumentor._azure_deployments = _PayiInstrumentor._model_mapping_to_context_dict(model_mappings)

    @staticmethod
    def instrument(instrumentor: _PayiInstrumentor) -> None:
        AnthropicInstrumentor._module_version = get_version_helper(AnthropicInstrumentor._module_name)

        wrappers = [
            ("anthropic._base_client", "AsyncAPIClient._process_response", _ProviderRequest.aprocess_response_wrapper),
            ("anthropic._base_client", "SyncAPIClient._process_response", _ProviderRequest.process_response_wrapper),
            ("anthropic.resources.messages", "Messages.create", messages_wrapper(instrumentor)),
            ("anthropic.resources.messages", "Messages.stream", stream_messages_wrapper(instrumentor)),
            ("anthropic.resources.beta.messages", "Messages.create", messages_wrapper(instrumentor)),
            ("anthropic.resources.beta.messages", "Messages.stream", stream_messages_wrapper(instrumentor)),
            ("anthropic.resources.messages", "AsyncMessages.create", amessages_wrapper(instrumentor)),
            ("anthropic.resources.messages", "AsyncMessages.stream", astream_messages_wrapper(instrumentor)),
            ("anthropic.resources.beta.messages", "AsyncMessages.create", amessages_wrapper(instrumentor)),
            ("anthropic.resources.beta.messages", "AsyncMessages.stream", astream_messages_wrapper(instrumentor)),

            # Wrap MessageStream iteration to track state across chunks
            ("anthropic.lib.streaming._messages", "MessageStream.__iter__", message_stream_iter_wrapper(instrumentor)),
            ("anthropic.lib.streaming._messages", "AsyncMessageStream.__aiter__", async_message_stream_aiter_wrapper(instrumentor)),
        ]

        for module, method, wrapper in wrappers:
            try:
                wrap_function_wrapper(module, method, wrapper)
            except Exception as e:
                instrumentor._logger.debug(f"Failed to wrap {module}.{method}: {e}")

@_PayiInstrumentor.payi_wrapper
def messages_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("Anthropic messages wrapper")
    return instrumentor.invoke_wrapper(
        _AnthropicProviderRequest(instrumentor=instrumentor, streaming_type=_StreamingType.iterator, instance=instance),
        _IsStreaming.kwargs,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_wrapper
def stream_messages_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("Anthropic stream wrapper")
    return instrumentor.invoke_wrapper(
        _AnthropicProviderRequest(instrumentor=instrumentor, streaming_type=_StreamingType.stream_manager, instance=instance),
        _IsStreaming.true,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_awrapper
async def amessages_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("aync Anthropic messages wrapper")
    return await instrumentor.async_invoke_wrapper(
        _AnthropicProviderRequest(instrumentor=instrumentor, streaming_type=_StreamingType.iterator, instance=instance),
        _IsStreaming.kwargs,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_awrapper
async def astream_messages_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("aync Anthropic stream wrapper")
    return await instrumentor.async_invoke_wrapper(
        _AnthropicProviderRequest(instrumentor=instrumentor, streaming_type=_StreamingType.stream_manager, instance=instance),
        _IsStreaming.true,
        wrapped,
        instance,
        args,
        kwargs,
    )

@_PayiInstrumentor.payi_wrapper
def message_stream_iter_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("MessageStream.__iter__ wrapper")
    
    # Messages.stream wrapper is expected to set attr _payi_request so that context is propagated and tracked

    # Get the stored request and tracking info
    request = getattr(instance, '_payi_request', None)
    
    if not request:
        instrumentor._logger.debug("MessageStream.__iter__ - missing request, returning original")
        return wrapped(*args, **kwargs)
    
    # Call the original __iter__ to get the iterator
    original_iterator = wrapped(*args, **kwargs)
    
    # Wrap it with GeneratorWrapper to track state across chunks
    return _GeneratorWrapper(
        generator=original_iterator,
        instance=instance,
        instrumentor=instrumentor,
        request=request,
    )

@_PayiInstrumentor.payi_awrapper
async def async_message_stream_aiter_wrapper(
    instrumentor: _PayiInstrumentor,
    wrapped: Any,
    instance: Any,
    *args: Any,
    **kwargs: Any,
) -> Any:
    instrumentor._logger.debug("AsyncMessageStream.__aiter__ wrapper")
    
    # AsyncMessages.stream wrapper is expected to set attr _payi_request so that context is propagated and tracked

    # Get the stored request and tracking info
    request = getattr(instance, '_payi_request', None)
    
    if not request:
        instrumentor._logger.debug("AsyncMessageStream.__aiter__ - missing request, returning original")
        return wrapped(*args, **kwargs)
    
    # Call the original __aiter__ to get the async iterator
    original_iterator = wrapped(*args, **kwargs)
    
    # Wrap it with GeneratorWrapper to track state across chunks
    return _GeneratorWrapper(
        generator=original_iterator,
        instance=instance,
        instrumentor=instrumentor,
        request=request,
    )

class _AnthropicProviderRequest(_ProviderRequest):
    def __init__(self, instrumentor: _PayiInstrumentor, streaming_type: _StreamingType, instance: Any = None) -> None:
        self._anthropic_client = instance._client if instance and hasattr(instance, "_client") else None
        self._is_vertex: bool = AnthropicInstrumentor.is_vertex(self._anthropic_client)
        self._is_bedrock: bool = AnthropicInstrumentor.is_bedrock(self._anthropic_client)
        self._is_azure: bool = AnthropicInstrumentor.is_azure(self._anthropic_client)
        self._is_anthropic_saas: bool = False

        category: str = ""
        if self._is_vertex:
            category = PayiCategories.google_vertex
        elif self._is_bedrock:
            category = PayiCategories.aws_bedrock
        elif self._is_azure:
            category = PayiCategories.azure
        else:
            category = PayiCategories.anthropic
            self._is_anthropic_saas = True

        instrumentor._logger.debug(f"Anthropic messages instrumenting category {category}")

        super().__init__(
            instrumentor=instrumentor,
            category=category,
            streaming_type=streaming_type,
            module_name=AnthropicInstrumentor._module_name,
            module_version=AnthropicInstrumentor._module_version,
            )

        if hasattr(self._anthropic_client, "base_url"):
           try:
               self.provider_uri = str(self._anthropic_client.base_url) # type: ignore
           except Exception:
               pass

    @override
    def process_chunk(self, chunk: Any) -> _ChunkResult:
        return anthropic_process_chunk(self, chunk.to_dict(), assign_id=True)

    @override
    def process_synchronous_response(self, response: Any, kwargs: Any) -> Any:
        anthropic_process_synchronous_response(
            request=self,
            response=response.to_dict(),
            log_prompt_and_response=self._log_prompt_and_response,
            assign_id=True)

        return None

    def _update_resource_name(self, model: str) -> str:
        return ("anthropic." if self._is_vertex else "") + model

    @override
    def process_request(self, instance: Any, args: Sequence[Any], kwargs: Any) -> bool:
        model = self._update_resource_name(kwargs.get("model", ""))
        self._ingest["resource"] = model

        if not self._price_as.resource and not self._price_as.category and AnthropicInstrumentor._azure_deployments:
            deployment = AnthropicInstrumentor._azure_deployments.get(model, {})
            self._price_as.category = deployment.get("price_as_category", None)
            self._price_as.resource = deployment.get("price_as_resource", None)
            self._price_as.resource_scope = deployment.get("resource_scope", None)

        if self._is_azure and not self._price_as.resource and not self._price_as.category:
            self._instrumentor._logger.debug(f"Azure Anthropic model {model}, available mappings: {list(AnthropicInstrumentor._azure_deployments.keys())}")
            self._instrumentor._logger.warning("Azure Anthropic requires price as resource and/or category to be specified unless mapped in the Pay-i service")

        if self._price_as.resource_scope:
            self._ingest["resource_scope"] = self._price_as.resource_scope
        
        # override defaults
        if self._price_as.category:
            self._ingest["category"] = self._price_as.category
        if self._price_as.resource:
            self._ingest["resource"] = self._update_resource_name(self._price_as.resource)

        self._instrumentor._logger.debug(f"Processing anthropic request: model {self._ingest['resource']}, category {self._category}")

        messages = kwargs.get("messages")
        if messages:
            anthropic_has_image_and_get_texts(self, messages)

        return True

    @override
    def remove_prompt_inline_data(self, prompt: 'dict[str, Any]') -> bool:
        return anthropic_remove_inline_data(prompt)

    @override
    def process_exception(self, exception: Exception, kwargs: Any, ) -> bool:
        try:
            status_code: Optional[int] = None

            if hasattr(exception, "status_code"):
                status_code = getattr(exception, "status_code", None)
                if isinstance(status_code, int):
                    self._ingest["http_status_code"] = status_code

            if not status_code:
                self.exception_to_semantic_failure(exception,)
                return True

            if hasattr(exception, "request_id"):
                request_id = getattr(exception, "request_id", None)
                if isinstance(request_id, str):
                    self._ingest["provider_response_id"] = request_id

            if hasattr(exception, "response"):
                response = getattr(exception, "response", None)
                if hasattr(response, "text"):
                    text = getattr(response, "text", None)
                    if isinstance(text, str):
                        self._ingest["provider_response_json"] = text

        except Exception as e:
            self._instrumentor._logger.debug(f"Error processing exception: {e}")
            return False

        return True

def anthropic_process_compute_input_cost(request: _ProviderRequest, usage: 'dict[str, Any]') -> int:
    input = usage.get('input_tokens', 0)
    units: dict[str, IngestUnits] = request._ingest["units"]

    cache_creation_input_tokens = usage.get("cache_creation_input_tokens", 0)
    cache_read_input_tokens = usage.get("cache_read_input_tokens", 0)

    total_input_tokens = input + cache_creation_input_tokens + cache_read_input_tokens

    request._is_large_context = total_input_tokens >= 200000
    large_context = "_large_context" if request._is_large_context else ""

    cache_creation: dict[str, int] = usage.get("cache_creation", {})
    ephemeral_5m_input_tokens: Optional[int] = None
    ephemeral_1h_input_tokens: Optional[int] = None
    textCacheWriteAdded = False

    if cache_creation:
        ephemeral_5m_input_tokens = cache_creation.get("ephemeral_5m_input_tokens", 0)
        if ephemeral_5m_input_tokens > 0:
            textCacheWriteAdded = True
            units["text_cache_write"+large_context] = IngestUnits(input=ephemeral_5m_input_tokens, output=0)

        ephemeral_1h_input_tokens = cache_creation.get("ephemeral_1h_input_tokens", 0)
        if ephemeral_1h_input_tokens > 0:
            textCacheWriteAdded = True
            units["text_cache_write_1h"+large_context] = IngestUnits(input=ephemeral_1h_input_tokens, output=0)

    if textCacheWriteAdded is False and cache_creation_input_tokens > 0:
        units["text_cache_write"+large_context] = IngestUnits(input=cache_creation_input_tokens, output=0)

    cache_read_input_tokens = usage.get("cache_read_input_tokens", 0)
    if cache_read_input_tokens > 0:
        units["text_cache_read"+large_context] = IngestUnits(input=cache_read_input_tokens, output=0)

    return request.update_for_vision(input)

def process_inference_geo(request: _ProviderRequest, inference_geo: Optional[str]) -> None:
    if not isinstance(request, _AnthropicProviderRequest):
        return

    anthropic_request: _AnthropicProviderRequest = request
    if not anthropic_request._is_anthropic_saas:
        return

    if inference_geo and inference_geo not in ('global', 'not_available'):
        request._ingest["resource_scope"] = f'{PayiResourceScopes.datazone_scope}.{inference_geo}'

def anthropic_process_synchronous_response(request: _ProviderRequest, response: 'dict[str, Any]', log_prompt_and_response: bool, assign_id: bool) -> Any:
    usage = response.get('usage', {})
    units: dict[str, IngestUnits] = request._ingest["units"]

    input_tokens = anthropic_process_compute_input_cost(request, usage)
    output = usage.get('output_tokens', 0)

    large_context = "_large_context" if request._is_large_context else ""
    units["text"+large_context] = IngestUnits(input=input_tokens, output=output)

    process_inference_geo(request, usage.get("inference_geo", ''))

    content = response.get('content', [])
    if content:
        for c in content:
            if c.get("type", "") != "tool_use":
                continue
            name = c.get("name", "")
            input = c.get("input", "")
            arguments: Optional[str] = None
            if input and isinstance(input, dict):
                arguments = json.dumps(input, ensure_ascii=False)
            
            if name and arguments:
                request.add_synchronous_function_call(name=name, arguments=arguments)

    if log_prompt_and_response:
        request._ingest["provider_response_json"] = json.dumps(response)
    
    if assign_id:
        request._ingest["provider_response_id"] = response.get('id', None)
    
    web_search_requests = usage.get("server_tool_use", {}).get("web_search_requests", 0)
    if web_search_requests > 0:
        units["web_search_request"] = IngestUnits(output=web_search_requests)

    return None

def anthropic_process_chunk(request: _ProviderRequest, chunk: 'dict[str, Any]', assign_id: bool) -> _ChunkResult:    
    ingest = False
    type = chunk.get('type', "")

    if type == "message_start":
        message = chunk['message']

        if assign_id:
            request._ingest["provider_response_id"] = message.get('id', None)

        model = message.get('model', None)
        if model and 'resource' in request._ingest:
            request._instrumentor._logger.debug(f"Anthropic streaming, reported model: {model}, instrumented model {request._ingest['resource']}")

        usage = message.get('usage', {})
        units = request._ingest["units"]

        input = anthropic_process_compute_input_cost(request, usage)

        large_context = "_large_context" if request._is_large_context else ""
        units["text"+large_context] = IngestUnits(input=input, output=0)

        process_inference_geo(request, usage.get("inference_geo", ''))

        request._instrumentor._logger.debug(f"Anthropic streaming captured {input} input tokens, ")

    elif type == "message_delta":
        usage = chunk.get('usage', {})
        ingest = True
        large_context = "_large_context" if request._is_large_context else ""

        # Web search will return an updated input tokens value at the end of streaming
        input_tokens = usage.get('input_tokens', None)
        if input_tokens is not None:
            request._instrumentor._logger.debug(f"Anthropic streaming finished, updated input tokens: {input_tokens}")
            request._ingest["units"]["text"+large_context]["input"] = input_tokens

        request._ingest["units"]["text"+large_context]["output"] = usage.get('output_tokens', 0)
        
        web_search_requests = usage.get("server_tool_use", {}).get("web_search_requests", 0)
        if web_search_requests > 0:
            request._ingest["units"]["web_search_request"] = IngestUnits(output=web_search_requests)

        request._instrumentor._logger.debug(f"Anthropic streaming finished: output tokens {usage.get('output_tokens', 0)} ")

    elif type == "content_block_start":
        request._building_function_response = False

        content_block = chunk.get('content_block', {})
        if content_block and content_block.get('type', "") == "tool_use":
            index = chunk.get('index', None)
            name = content_block.get('name', "")

            if index and isinstance(index, int) and name:
                request._building_function_response = True
                request.add_streaming_function_call(index=index, name=name, arguments=None)

    elif type == "content_block_delta":
        if request._building_function_response:
            delta = chunk.get("delta", {})
            type = delta.get("type", "")
            partial_json = delta.get("partial_json", "")
            index = chunk.get('index', None)

            if index and isinstance(index, int) and type == "input_json_delta" and partial_json:
                request.add_streaming_function_call(index=index, name=None, arguments=partial_json)

    elif type == "content_block_stop":
        request._building_function_response = False

    else:
        request._instrumentor._logger.debug(f"Anthropic streaming chunk: {type}")
        
    return _ChunkResult(send_chunk_to_caller=True, ingest=ingest)

def anthropic_has_image_and_get_texts(request: _ProviderRequest, messages: Any) -> None:
    estimated_token_count = 0 
    has_image = False

    try:
        enc = tiktoken.get_encoding("cl100k_base")
        for message in messages:
            msg_has_image, msg_prompt_tokens = has_image_and_get_texts(enc, message.get('content', ''))
            if msg_has_image:
                has_image = True
                estimated_token_count += msg_prompt_tokens
        
        if has_image and estimated_token_count > 0:
            request._estimated_prompt_tokens = estimated_token_count

    except Exception:
        request._instrumentor._logger.info("Anthropic skipping vision token calc, could not load cl100k_base")

def has_image_and_get_texts(encoding: tiktoken.Encoding, content: Union[str, 'list[Any]']) -> 'tuple[bool, int]':
    if isinstance(content, list): # type: ignore
        has_image = any(item.get("type") == "image" for item in content)
        if has_image is False:
            return has_image, 0
        
        token_count = sum(len(encoding.encode(item.get("text", ""))) for item in content if item.get("type") == "text")
        return has_image, token_count
    
    return False, 0

def anthropic_remove_inline_data(prompt: 'dict[str, Any]') -> bool:# noqa: ARG002
    messages = prompt.get("messages", [])
    if not messages:
        return False

    modified = False
    for message in messages:
        content = message.get('content', Any)
        if not content or not isinstance(content, list):
            continue

        for item in content: # type: ignore
            if not isinstance(item, dict):
                continue
            # item: dict[str, Any]
            type = item.get("type", "") # type: ignore
            if type != "image":
                continue

            source = item.get("source", {}) # type: ignore
            if source.get("type", "") == "base64": # type: ignore
                source["data"] = _PayiInstrumentor._not_instrumented
                modified = True

    return modified



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional

from pydantic import Field as FieldInfo

from ..._models import BaseModel
from .xproxy_error import XproxyError

__all__ = ["APIError"]


class APIError(BaseModel):
    message: str

    status_code: int = FieldInfo(alias="statusCode")

    xproxy_error: Optional[XproxyError] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional
from datetime import datetime

from .._models import BaseModel
from .shared.api_error import APIError

__all__ = ["BulkIngestResponse", "Error"]


class Error(BaseModel):
    item_index: Optional[int] = None

    xproxy_result: Optional[APIError] = None


class BulkIngestResponse(BaseModel):
    ingest_count: int

    ingest_timestamp: datetime

    request_id: str

    error_count: Optional[int] = None

    errors: Optional[List[Error]] = None

    item_error_count: Optional[int] = None

    total_count: Optional[int] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .limits import (
    LimitsResource,
    AsyncLimitsResource,
    LimitsResourceWithRawResponse,
    AsyncLimitsResourceWithRawResponse,
    LimitsResourceWithStreamingResponse,
    AsyncLimitsResourceWithStreamingResponse,
)
from .properties import (
    PropertiesResource,
    AsyncPropertiesResource,
    PropertiesResourceWithRawResponse,
    AsyncPropertiesResourceWithRawResponse,
    PropertiesResourceWithStreamingResponse,
    AsyncPropertiesResourceWithStreamingResponse,
)

__all__ = [
    "PropertiesResource",
    "AsyncPropertiesResource",
    "PropertiesResourceWithRawResponse",
    "AsyncPropertiesResourceWithRawResponse",
    "PropertiesResourceWithStreamingResponse",
    "AsyncPropertiesResourceWithStreamingResponse",
    "LimitsResource",
    "AsyncLimitsResource",
    "LimitsResourceWithRawResponse",
    "AsyncLimitsResourceWithRawResponse",
    "LimitsResourceWithStreamingResponse",
    "AsyncLimitsResourceWithStreamingResponse",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Iterable, Optional
from datetime import datetime
from typing_extensions import Required, Annotated, TypedDict

from .._types import SequenceNotStr
from .._utils import PropertyInfo
from .shared_params.ingest_units import IngestUnits
from .pay_i_common_models_api_router_header_info_param import PayICommonModelsAPIRouterHeaderInfoParam

__all__ = ["IngestUnitsParams", "ProviderResponseFunctionCall"]


class IngestUnitsParams(TypedDict, total=False):
    category: Required[str]

    units: Required[Dict[str, IngestUnits]]

    end_to_end_latency_ms: Optional[int]

    event_timestamp: Annotated[Union[str, datetime, None], PropertyInfo(format="iso8601")]

    http_status_code: Optional[int]

    properties: Optional[Dict[str, Optional[str]]]

    provider_request_headers: Optional[Iterable[PayICommonModelsAPIRouterHeaderInfoParam]]

    provider_request_json: Optional[str]

    provider_request_reasoning_json: Optional[str]

    provider_response_function_calls: Optional[Iterable[ProviderResponseFunctionCall]]

    provider_response_headers: Optional[Iterable[PayICommonModelsAPIRouterHeaderInfoParam]]

    provider_response_id: Optional[str]

    provider_response_json: Union[str, SequenceNotStr[str], None]

    provider_uri: Optional[str]

    resource: Optional[str]

    time_to_first_completion_token_ms: Optional[int]

    time_to_first_token_ms: Optional[int]

    use_case_properties: Optional[Dict[str, Optional[str]]]

    limit_ids: Annotated[Union[list[str], None], PropertyInfo(alias="xProxy-Limit-IDs")]

    request_tags: Annotated[Union[list[str], None], PropertyInfo(alias="xProxy-Request-Tags")]

    use_case_name: Annotated[Union[str, None], PropertyInfo(alias="xProxy-UseCase-Name")]

    use_case_id: Annotated[Union[str, None], PropertyInfo(alias="xProxy-UseCase-ID")]

    use_case_version: Annotated[Union[int, None], PropertyInfo(alias="xProxy-UseCase-Version")]

    resource_scope: Annotated[Union[str, None], PropertyInfo(alias="xProxy-Resource-Scope")]

    use_case_step: Annotated[Union[str, None], PropertyInfo(alias="xProxy-UseCase-Step")]

    user_id: Annotated[Union[str, None], PropertyInfo(alias="xProxy-User-ID")]

    account_name: Annotated[str, PropertyInfo(alias="xProxy-Account-Name")]

    disable_logging: Optional[bool]

class ProviderResponseFunctionCall(TypedDict, total=False):
    name: Required[str]

    arguments: Optional[str]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .requests import (
    RequestsResource,
    AsyncRequestsResource,
    RequestsResourceWithRawResponse,
    AsyncRequestsResourceWithRawResponse,
    RequestsResourceWithStreamingResponse,
    AsyncRequestsResourceWithStreamingResponse,
)
from .request_id import (
    RequestIDResource,
    AsyncRequestIDResource,
    RequestIDResourceWithRawResponse,
    AsyncRequestIDResourceWithRawResponse,
    RequestIDResourceWithStreamingResponse,
    AsyncRequestIDResourceWithStreamingResponse,
)
from .response_id import (
    ResponseIDResource,
    AsyncResponseIDResource,
    ResponseIDResourceWithRawResponse,
    AsyncResponseIDResourceWithRawResponse,
    ResponseIDResourceWithStreamingResponse,
    AsyncResponseIDResourceWithStreamingResponse,
)

__all__ = [
    "RequestIDResource",
    "AsyncRequestIDResource",
    "RequestIDResourceWithRawResponse",
    "AsyncRequestIDResourceWithRawResponse",
    "RequestIDResourceWithStreamingResponse",
    "AsyncRequestIDResourceWithStreamingResponse",
    "ResponseIDResource",
    "AsyncResponseIDResource",
    "ResponseIDResourceWithRawResponse",
    "AsyncResponseIDResourceWithRawResponse",
    "ResponseIDResourceWithStreamingResponse",
    "AsyncResponseIDResourceWithStreamingResponse",
    "RequestsResource",
    "AsyncRequestsResource",
    "RequestsResourceWithRawResponse",
    "AsyncRequestsResourceWithRawResponse",
    "RequestsResourceWithStreamingResponse",
    "AsyncRequestsResourceWithStreamingResponse",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Annotated, TypedDict

from .._utils import PropertyInfo

__all__ = ["CategoryListResourcesParams"]


class CategoryListResourcesParams(TypedDict, total=False):
    active: Annotated[bool, PropertyInfo(alias="Active")]

    cursor: str

    limit: int

    sort_ascending: bool



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from .request_id.request_id import (
    RequestIDResource,
    AsyncRequestIDResource,
    RequestIDResourceWithRawResponse,
    AsyncRequestIDResourceWithRawResponse,
    RequestIDResourceWithStreamingResponse,
    AsyncRequestIDResourceWithStreamingResponse,
)
from .response_id.response_id import (
    ResponseIDResource,
    AsyncResponseIDResource,
    ResponseIDResourceWithRawResponse,
    AsyncResponseIDResourceWithRawResponse,
    ResponseIDResourceWithStreamingResponse,
    AsyncResponseIDResourceWithStreamingResponse,
)

__all__ = ["RequestsResource", "AsyncRequestsResource"]


class RequestsResource(SyncAPIResource):
    @cached_property
    def request_id(self) -> RequestIDResource:
        return RequestIDResource(self._client)

    @cached_property
    def response_id(self) -> ResponseIDResource:
        return ResponseIDResource(self._client)

    @cached_property
    def with_raw_response(self) -> RequestsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return RequestsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> RequestsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return RequestsResourceWithStreamingResponse(self)


class AsyncRequestsResource(AsyncAPIResource):
    @cached_property
    def request_id(self) -> AsyncRequestIDResource:
        return AsyncRequestIDResource(self._client)

    @cached_property
    def response_id(self) -> AsyncResponseIDResource:
        return AsyncResponseIDResource(self._client)

    @cached_property
    def with_raw_response(self) -> AsyncRequestsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return AsyncRequestsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncRequestsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return AsyncRequestsResourceWithStreamingResponse(self)


class RequestsResourceWithRawResponse:
    def __init__(self, requests: RequestsResource) -> None:
        self._requests = requests

    @cached_property
    def request_id(self) -> RequestIDResourceWithRawResponse:
        return RequestIDResourceWithRawResponse(self._requests.request_id)

    @cached_property
    def response_id(self) -> ResponseIDResourceWithRawResponse:
        return ResponseIDResourceWithRawResponse(self._requests.response_id)


class AsyncRequestsResourceWithRawResponse:
    def __init__(self, requests: AsyncRequestsResource) -> None:
        self._requests = requests

    @cached_property
    def request_id(self) -> AsyncRequestIDResourceWithRawResponse:
        return AsyncRequestIDResourceWithRawResponse(self._requests.request_id)

    @cached_property
    def response_id(self) -> AsyncResponseIDResourceWithRawResponse:
        return AsyncResponseIDResourceWithRawResponse(self._requests.response_id)


class RequestsResourceWithStreamingResponse:
    def __init__(self, requests: RequestsResource) -> None:
        self._requests = requests

    @cached_property
    def request_id(self) -> RequestIDResourceWithStreamingResponse:
        return RequestIDResourceWithStreamingResponse(self._requests.request_id)

    @cached_property
    def response_id(self) -> ResponseIDResourceWithStreamingResponse:
        return ResponseIDResourceWithStreamingResponse(self._requests.response_id)


class AsyncRequestsResourceWithStreamingResponse:
    def __init__(self, requests: AsyncRequestsResource) -> None:
        self._requests = requests

    @cached_property
    def request_id(self) -> AsyncRequestIDResourceWithStreamingResponse:
        return AsyncRequestIDResourceWithStreamingResponse(self._requests.request_id)

    @cached_property
    def response_id(self) -> AsyncResponseIDResourceWithStreamingResponse:
        return AsyncResponseIDResourceWithStreamingResponse(self._requests.response_id)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .api_error import APIError as APIError
from .ingest_units import IngestUnits as IngestUnits
from .xproxy_error import XproxyError as XproxyError
from .xproxy_result import XproxyResult as XproxyResult
from .properties_request import PropertiesRequest as PropertiesRequest
from .properties_response import PropertiesResponse as PropertiesResponse
from .pay_i_common_models_budget_management_cost_details_base import (
    PayICommonModelsBudgetManagementCostDetailsBase as PayICommonModelsBudgetManagementCostDetailsBase,
)
from .pay_i_common_models_budget_management_create_limit_base import (
    PayICommonModelsBudgetManagementCreateLimitBase as PayICommonModelsBudgetManagementCreateLimitBase,
)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Iterable, Optional
from datetime import datetime

import httpx

from payi._utils._utils import is_given

from ..types import ingest_units_params
from .._types import Body, Omit, Query, Headers, NotGiven, SequenceNotStr, omit, not_given
from .._utils import maybe_transform, strip_not_given, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.ingest_response import IngestResponse
from ..types.bulk_ingest_response import BulkIngestResponse
from ..types.bulk_ingest_request_param import BulkIngestRequestParam
from ..types.shared_params.ingest_units import IngestUnits
from ..types.pay_i_common_models_api_router_header_info_param import PayICommonModelsAPIRouterHeaderInfoParam

__all__ = ["IngestResource", "AsyncIngestResource"]

def convert_property_values_to_str(properties: Dict[str, Optional[str]]) -> Dict[str, Optional[str]]:
    converted_properties: Dict[str, Optional[str]] = {}
    for k, v in properties.items():
        if v is None:
            converted_properties[k] = None
        else:
            try:
                converted_properties[k] = str(v)
            except Exception:
                pass  # Skip this key-value pair if str() fails

    return converted_properties

class IngestResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> IngestResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return IngestResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> IngestResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return IngestResourceWithStreamingResponse(self)

    def bulk(
        self,
        *,
        events: Iterable[BulkIngestRequestParam] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BulkIngestResponse:
        """
        Bulk Ingest

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/api/v1/ingest/bulk",
            body=maybe_transform(events, Iterable[BulkIngestRequestParam]),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BulkIngestResponse,
        )

    def units(
        self,
        *,
        category: str,
        units: Dict[str, IngestUnits],
        end_to_end_latency_ms: Optional[int] | Omit = omit,
        event_timestamp: Union[str, datetime, None] | Omit = omit,
        http_status_code: Optional[int] | Omit = omit,
        properties: Optional[Dict[str, Optional[str]]] | Omit = omit,
        provider_request_headers: Optional[Iterable[PayICommonModelsAPIRouterHeaderInfoParam]] | Omit = omit,
        provider_request_json: Optional[str] | Omit = omit,
        provider_request_reasoning_json: Optional[str] | Omit = omit,
        provider_response_function_calls: Optional[Iterable[ingest_units_params.ProviderResponseFunctionCall]]
        | Omit = omit,
        provider_response_headers: Optional[Iterable[PayICommonModelsAPIRouterHeaderInfoParam]] | Omit = omit,
        provider_response_id: Optional[str] | Omit = omit,
        provider_response_json: Union[str, SequenceNotStr[str], None] | Omit = omit,
        provider_uri: Optional[str] | Omit = omit,
        resource: Optional[str] | Omit = omit,
        time_to_first_completion_token_ms: Optional[int] | Omit = omit,
        time_to_first_token_ms: Optional[int] | Omit = omit,
        use_case_properties: Optional[Dict[str, Optional[str]]] | Omit = omit,
        limit_ids: Optional[list[str]] | Omit = omit,
        disable_logging: Optional[bool] | Omit = omit,
        request_tags: Optional[list[str]] | Omit = omit,
        use_case_id: Optional[str] | Omit = omit,
        use_case_name: Optional[str] | Omit = omit,
        use_case_step: Optional[str] | Omit = omit,
        use_case_version: Optional[int] | Omit = omit,
        user_id: Optional[str] | Omit = omit,
        resource_scope: Optional[str] | Omit = omit,
        account_name: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> IngestResponse:
        """
        Ingest an Event

        Args:
          category (str): The name of the category

          resource (str): The name of the resource

          input (int): The number of input units

          output (int): The number of output units

          event_timestamp: (str, datetime, None): The timestamp of the event

          disable_logging (bool, optional): Disable logging for the request

          limit_ids (list[str], optional): The limit IDs to associate with the request

          properties (Dict[str, Optional[str]], optional): Properties to associate with the request

          request_tags (list[str], optional): The request tags to associate with the request

          use_case_name (str, optional): The use case name

          use_case_id (str, optional): The use case instance id

          use_case_step (str, optional): The use case step

          use_case_version (int, optional): The use case instance version

          use_case_properties (Dict[str, Optional[str]], optional): The use case properties

          user_id (str, optional): The user id
          
          resource_scope(str, optional): The scope of the resource

          account_name (str, optional): The account name

          extra_headers (Dict[str, str], optional): Additional headers for the request. Defaults to None.

          extra_query (Dict[str, str], optional): Additional query parameters. Defaults to None.

          extra_body (Dict[str, Any], optional): Additional body parameters. Defaults to None.

          timeout (Union[float, None], optional): The timeout for the request in seconds. Defaults to None.
        """
        request_tags = request_tags
        valid_ids_str: str | Omit = omit
        use_case_version_str: str | Omit = omit

        if limit_ids is None or not is_given(limit_ids):
            valid_ids_str = omit
        elif not isinstance(limit_ids, list):  # type: ignore
            raise TypeError("limit_ids must be a list")
        else:
            # Proceed with the list comprehension if limit_ids is given
            valid_ids = [id.strip() for id in limit_ids if id.strip()]
            valid_ids_str = ",".join(valid_ids) if valid_ids else omit

        if use_case_name is None or not is_given(use_case_name):
            use_case_name = omit

        if use_case_step is None or not is_given(use_case_step):
            use_case_step = omit

        if use_case_id is None or not is_given(use_case_id):
            use_case_id = omit

        if use_case_version is None or not is_given(use_case_version):
            use_case_version_str = omit
        else:
            use_case_version_str = str(use_case_version)

        if use_case_properties and is_given(use_case_properties):
            use_case_properties = convert_property_values_to_str(use_case_properties)

        if user_id is None or not is_given(user_id):
            user_id = omit

        if resource_scope is None or not is_given(resource_scope):
            resource_scope = omit

        if account_name is None or not is_given(account_name):
            account_name = omit

        if properties and is_given(properties):
            properties = convert_property_values_to_str(properties)
        
        extra_headers = {
            **strip_not_given(
                {
                    "xProxy-Limit-IDs": valid_ids_str,
                    "xProxy-Request-Tags": omit,
                    "xProxy-UseCase-ID": use_case_id,
                    "xProxy-UseCase-Name": use_case_name,
                    "xProxy-UseCase-Step": use_case_step,
                    "xProxy-UseCase-Version": use_case_version_str
                    if is_given(use_case_version)
                    else not_given,
                    "xProxy-User-ID": user_id,
                    "xProxy-Resource-Scope": resource_scope,
                    "xProxy-Account-Name": account_name,
                    "xProxy-Logging-Disable": str(disable_logging)
                    if is_given(disable_logging)
                    else not_given,
                }
            ),
            **(extra_headers or {}),
        }

        return self._post(
            "/api/v1/ingest",
            body=maybe_transform(
                {
                    "category": category,
                    "units": units,
                    "end_to_end_latency_ms": end_to_end_latency_ms,
                    "event_timestamp": event_timestamp,
                    "http_status_code": http_status_code,
                    "properties": properties,
                    "provider_request_headers": provider_request_headers,
                    "provider_request_json": provider_request_json,
                    "provider_request_reasoning_json": provider_request_reasoning_json,
                    "provider_response_function_calls": provider_response_function_calls,
                    "provider_response_headers": provider_response_headers,
                    "provider_response_id": provider_response_id,
                    "provider_response_json": provider_response_json,
                    "provider_uri": provider_uri,
                    "resource": resource,
                    "time_to_first_completion_token_ms": time_to_first_completion_token_ms,
                    "time_to_first_token_ms": time_to_first_token_ms,
                    "use_case_properties": use_case_properties,
                },
                ingest_units_params.IngestUnitsParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=IngestResponse,
        )


class AsyncIngestResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncIngestResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/Pay-i/pay-i-python#accessing-raw-response-data-eg-headers
        """
        return AsyncIngestResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncIngestResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/Pay-i/pay-i-python#with_streaming_response
        """
        return AsyncIngestResourceWithStreamingResponse(self)

    async def bulk(
        self,
        *,
        events: Iterable[BulkIngestRequestParam] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BulkIngestResponse:
        """
        Bulk Ingest

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/api/v1/ingest/bulk",
            body=await async_maybe_transform(events, Iterable[BulkIngestRequestParam]),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BulkIngestResponse,
        )

    async def units(
        self,
        *,
        category: str,
        units: Dict[str, IngestUnits],
        end_to_end_latency_ms: Optional[int] | Omit = omit,
        event_timestamp: Union[str, datetime, None] | Omit = omit,
        http_status_code: Optional[int] | Omit = omit,
        properties: Optional[Dict[str, Optional[str]]] | Omit = omit,
        provider_request_headers: Optional[Iterable[PayICommonModelsAPIRouterHeaderInfoParam]] | Omit = omit,
        provider_request_json: Optional[str] | Omit = omit,
        provider_request_reasoning_json: Optional[str] | Omit = omit,
        provider_response_function_calls: Optional[Iterable[ingest_units_params.ProviderResponseFunctionCall]]
        | Omit = omit,
        provider_response_headers: Optional[Iterable[PayICommonModelsAPIRouterHeaderInfoParam]] | Omit = omit,
        provider_response_id: Optional[str] | Omit = omit,
        provider_response_json: Union[str, SequenceNotStr[str], None] | Omit = omit,
        provider_uri: Optional[str] | Omit = omit,
        resource: Optional[str] | Omit = omit,
        time_to_first_completion_token_ms: Optional[int] | Omit = omit,
        time_to_first_token_ms: Optional[int] | Omit = omit,
        use_case_properties: Optional[Dict[str, Optional[str]]] | Omit = omit,
        limit_ids: Optional[list[str]] | Omit = omit,
        disable_logging: Optional[bool] | Omit = omit,
        request_tags: Optional[list[str]] | Omit = omit,
        use_case_id: Optional[str] | Omit = omit,
        use_case_name: Optional[str] | Omit = omit,
        use_case_step: Optional[str] | Omit = omit,
        use_case_version: Optional[int] | Omit = omit,
        user_id: Optional[str] | Omit = omit,
        resource_scope: Union[str, None] | Omit = omit,
        account_name: Optional[str] | Omit = omit,
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> IngestResponse:
        """
        Ingest an Event

        Args:
          category (str): The name of the category

          resource (str): The name of the resource

          input (int): The number of input units

          output (int): The number of output units

          event_timestamp: (datetime, None): The timestamp of the event

          disable_logging: (bool, optional): Disable logging for the request
          
          limit_ids (list[str], optional): The limit IDs to associate with the request 

          properties (Dict[str, Optional[str]], optional): Properties to associate with the request

          request_tags (list[str], optional): The request tags to associate with the request

          use_case_name (str, optional): The use case name

          use_case_step (str, optional): The use case step

          use_case_id (str, optional): The use case instance id

          use_case_version (int, optional): The use case instance version

          use_case_properties (Dict[str, Optional[str]], optional): The use case properties

          user_id (str, optional): The user id
          
          resource_scope (str, optional): The scope of the resource

          account_name (str, optional): The account name

          extra_headers (Dict[str, str], optional): Additional headers for the request. Defaults to None.

          extra_query (Dict[str, str], optional): Additional query parameters. Defaults to None.

          extra_body (Dict[str, Any], optional): Additional body parameters. Defaults to None.

          timeout (Union[float, None], optional): The timeout for the request in seconds. Defaults to None.
        """
        request_tags = request_tags
        valid_ids_str: str | Omit = omit
        use_case_version_str: str | Omit = omit

        if limit_ids is None or not is_given(limit_ids):
            valid_ids_str = omit
        elif not isinstance(limit_ids, list):  # type: ignore
            raise TypeError("limit_ids must be a list")
        else:
            # Proceed with the list comprehension if limit_ids is given
            valid_ids = [id.strip() for id in limit_ids if id.strip()]
            valid_ids_str = ",".join(valid_ids) if valid_ids else omit

        if use_case_name is None or not is_given(use_case_name):
            use_case_name = omit

        if use_case_step is None or not is_given(use_case_step):
            use_case_step = omit

        if use_case_id is None or not is_given(use_case_id):
            use_case_id = omit

        if use_case_version is None or not is_given(use_case_version):
            use_case_version_str = omit
        else:
            use_case_version_str = str(use_case_version)

        if use_case_properties and is_given(use_case_properties):
            use_case_properties = convert_property_values_to_str(use_case_properties)

        if user_id is None or not is_given(user_id):
            user_id = omit

        if resource_scope is None or not is_given(resource_scope):
            resource_scope = omit

        if account_name is None or not is_given(account_name):
            account_name = omit

        if properties and is_given(properties):
            properties = convert_property_values_to_str(properties)
        
        extra_headers = {
            **strip_not_given(
                {
                    "xProxy-Account-Name": account_name,
                    "xProxy-Limit-IDs": valid_ids_str,
                    "xProxy-Request-Tags": omit,
                    "xProxy-UseCase-ID": use_case_id,
                    "xProxy-UseCase-Name": use_case_name,
                    "xProxy-UseCase-Step": use_case_step,
                    "xProxy-UseCase-Version": use_case_version_str
                    if is_given(use_case_version)
                    else not_given,
                    "xProxy-User-ID": user_id,
                    "xProxy-Resource-Scope": resource_scope,
                    "xProxy-Logging-Disable": str(disable_logging)
                    if is_given(disable_logging)
                    else not_given,
                }
            ),
            **(extra_headers or {}),
        }
        return await self._post(
            "/api/v1/ingest",
            body=await async_maybe_transform(
                {
                    "category": category,
                    "units": units,
                    "end_to_end_latency_ms": end_to_end_latency_ms,
                    "event_timestamp": event_timestamp,
                    "http_status_code": http_status_code,
                    "properties": properties,
                    "provider_request_headers": provider_request_headers,
                    "provider_request_json": provider_request_json,
                    "provider_request_reasoning_json": provider_request_reasoning_json,
                    "provider_response_function_calls": provider_response_function_calls,
                    "provider_response_headers": provider_response_headers,
                    "provider_response_id": provider_response_id,
                    "provider_response_json": provider_response_json,
                    "provider_uri": provider_uri,
                    "resource": resource,
                    "time_to_first_completion_token_ms": time_to_first_completion_token_ms,
                    "time_to_first_token_ms": time_to_first_token_ms,
                    "use_case_properties": use_case_properties,
                },
                ingest_units_params.IngestUnitsParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=IngestResponse,
        )


class IngestResourceWithRawResponse:
    def __init__(self, ingest: IngestResource) -> None:
        self._ingest = ingest

        self.bulk = to_raw_response_wrapper(
            ingest.bulk,
        )
        self.units = to_raw_response_wrapper(
            ingest.units,
        )


class AsyncIngestResourceWithRawResponse:
    def __init__(self, ingest: AsyncIngestResource) -> None:
        self._ingest = ingest

        self.bulk = async_to_raw_response_wrapper(
            ingest.bulk,
        )
        self.units = async_to_raw_response_wrapper(
            ingest.units,
        )


class IngestResourceWithStreamingResponse:
    def __init__(self, ingest: IngestResource) -> None:
        self._ingest = ingest

        self.bulk = to_streamed_response_wrapper(
            ingest.bulk,
        )
        self.units = to_streamed_response_wrapper(
            ingest.units,
        )


class AsyncIngestResourceWithStreamingResponse:
    def __init__(self, ingest: AsyncIngestResource) -> None:
        self._ingest = ingest

        self.bulk = async_to_streamed_response_wrapper(
            ingest.bulk,
        )
        self.units = async_to_streamed_response_wrapper(
            ingest.units,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Optional

from ..._models import BaseModel

__all__ = ["PropertiesRequest"]


class PropertiesRequest(BaseModel):
    properties: Dict[str, Optional[str]]



from __future__ import annotations

import os
import json
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence
from functools import wraps
from typing_extensions import override

from wrapt import ObjectProxy, wrap_function_wrapper  # type: ignore

from payi.lib.helpers import PayiCategories, PayiHeaderNames, PayiPropertyNames, payi_aws_bedrock_url
from payi.types.ingest_units_params import IngestUnits

from .instrument import (
    PayiInstrumentAwsBedrockConfig,
    _Context,
    _IsStreaming,
    _PayiInstrumentor,
)
from .version_helper import get_version_helper
from .ProviderRequest import _ChunkResult, _StreamingType, _ProviderRequest

if TYPE_CHECKING:
    from tokenizers import Tokenizer  # type: ignore
else:
    Tokenizer = None  

GUARDRAIL_SEMANTIC_FAILURE_DESCRIPTION = "Bedrock Guardrails intervened"

class BedrockInstrumentor:
    _module_name: str = "boto3"
    _module_version: str = ""

    _instrumentor: _PayiInstrumentor

    _guardrail_trace: bool = True

    _model_mapping: Dict[str, _Context] = {}

    _add_streaming_xproxy_result: bool = False

    @staticmethod
    def get_mapping(model_id: Optional[str]) -> _Context:
        if not model_id:
            return  {}

        return BedrockInstrumentor._model_mapping.get(model_id, {})

    @staticmethod
    def configure(aws_config: Optional[PayiInstrumentAwsBedrockConfig]) -> None:
        if not aws_config:
            return
        
        trace = aws_config.get("guardrail_trace", True)
        if trace is None:
            trace = True
        BedrockInstrumentor._guardrail_trace = trace

        add_streaming_xproxy_result = aws_config.get("add_streaming_xproxy_result", False)
        if add_streaming_xproxy_result:
            BedrockInstrumentor._add_streaming_xproxy_result = add_streaming_xproxy_result

        model_mappings = aws_config.get("model_mappings", [])
        if model_mappings:
            BedrockInstrumentor._model_mapping = _PayiInstrumentor._model_mapping_to_context_dict(model_mappings)

    @staticmethod
    def instrument(instrumentor: _PayiInstrumentor) -> None:
        BedrockInstrumentor._instrumentor = instrumentor

        BedrockInstrumentor._module_version = get_version_helper(BedrockInstrumentor._module_name)

        try:
            wrap_function_wrapper(
                "botocore.client",
                "ClientCreator.create_client",
                create_client_wrapper(instrumentor),
            )

            wrap_function_wrapper(
                "botocore.session",
                "Session.create_client",
                create_client_wrapper(instrumentor),
            )

        except Exception as e:
            instrumentor._logger.debug(f"Error instrumenting bedrock: {e}")
            return

@_PayiInstrumentor.payi_wrapper
def create_client_wrapper(instrumentor: _PayiInstrumentor, wrapped: Any, instance: Any, *args: Any, **kwargs: Any) -> Any: #  noqa: ARG001
    if kwargs.get("service_name") != "bedrock-runtime":
        # instrumentor._logger.debug(f"skipping client wrapper creation for {kwargs.get('service_name', '')} service")
        return wrapped(*args, **kwargs)

    try:
        client: Any = wrapped(*args, **kwargs)
        client.invoke_model = wrap_invoke(instrumentor, client.invoke_model, client)
        client.invoke_model_with_response_stream = wrap_invoke_stream(instrumentor, client.invoke_model_with_response_stream, client)
        client.converse = wrap_converse(instrumentor, client.converse, client)
        client.converse_stream = wrap_converse_stream(instrumentor, client.converse_stream, client)

        instrumentor._logger.debug(f"Instrumented bedrock client")

        if BedrockInstrumentor._instrumentor._proxy_default:
            # Register client callbacks to handle the Pay-i extra_headers parameter in the inference calls and redirect the request to the Pay-i endpoint
            _register_bedrock_client_callbacks(client)
            instrumentor._logger.debug(f"Registered bedrock client callbaks for proxy")

        return client
    except Exception as e:
        instrumentor._logger.debug(f"Error instrumenting bedrock client: {e}")
    
    return wrapped(*args, **kwargs)

BEDROCK_REQUEST_NAMES = [
    'request-created.bedrock-runtime.Converse',
    'request-created.bedrock-runtime.ConverseStream',
    'request-created.bedrock-runtime.InvokeModel',
    'request-created.bedrock-runtime.InvokeModelWithResponseStream',
]

def _register_bedrock_client_callbacks(client: Any) -> None:
    # Pass a unqiue_id to avoid registering the same callback multiple times in case this cell executed more than once
    # Redirect the request to the Pay-i endpoint after the request has been signed. 
    client.meta.events.register_last('request-created', _redirect_to_payi, unique_id=_redirect_to_payi)

def _redirect_to_payi(request: Any, event_name: str, **_: 'dict[str, Any]') -> None:
    from urllib3.util import parse_url
    from urllib3.util.url import Url

    if not event_name in BEDROCK_REQUEST_NAMES:
        return
    
    parsed_url: Url = parse_url(request.url)
    route_path = parsed_url.path
    request.url = f"{payi_aws_bedrock_url()}{route_path}"

    request.headers[PayiHeaderNames.api_key] = os.environ.get("PAYI_API_KEY", "")
    request.headers[PayiHeaderNames.provider_base_uri] = parsed_url.scheme + "://" + parsed_url.host # type: ignore
    
    extra_headers = BedrockInstrumentor._instrumentor._create_extra_headers()

    for key, value in extra_headers.items():
        request.headers[key] = value

class InvokeResponseWrapper(ObjectProxy): # type: ignore
    _cohere_embed_english_v3_tokenizer: Optional['Tokenizer'] = None

    def __init__(
        self,
        response: 'dict[str, Any]',
        body: Any,
        request: '_BedrockInvokeProviderRequest',
        ) -> None:

        super().__init__(body) # type: ignore
        self._response = response
        self._body = body
        self._request = request
        self._log_prompt_and_response = request._log_prompt_and_response

    def read(self, amt: Any =None) -> Any: # type: ignore
        # data is array of bytes
        data: bytes = self.__wrapped__.read(amt) # type: ignore
        response = json.loads(data) # type: ignore

        ingest = self._request._ingest

        # resource = ingest["resource"]
        # if not resource:
        #     return
        
        input: int = 0
        output: int = 0
        units: dict[str, IngestUnits] = ingest["units"]

        if self._request._is_anthropic:
            from .AnthropicInstrumentor import anthropic_process_synchronous_response

            anthropic_process_synchronous_response(
                request=self._request, 
                response=response,
                log_prompt_and_response=False, # will evaluate logging later
                assign_id=False)

        elif self._request._is_meta:
            input = response.get('prompt_token_count', 0)
            output = response.get('generation_token_count', 0)
            units["text"] = IngestUnits(input=input, output=output)

        elif self._request._is_nova:
            usage = response.get("usage", {})

            input = usage.get("inputTokens", 0)
            output = usage.get("outputTokens", 0)
            units["text"] = IngestUnits(input=input, output=output)

            text_cache_read = usage.get("cacheReadInputTokenCount", None)
            if text_cache_read:
                units["text_cache_read"] = text_cache_read

            text_cache_write = usage.get("cacheWriteInputTokenCount", None)
            if text_cache_write:
                units["text_cache_write"] = text_cache_write

            bedrock_converse_process_synchronous_function_call(self._request, response)

        elif self._request._is_amazon_titan_embed_text_v1:
            input = response.get('inputTextTokenCount', 0)
            units["text"] = IngestUnits(input=input, output=0)

        elif self._request._is_cohere_embed_english_v3:
            texts: list[str] = response.get("texts", [])
            if texts and len(texts) > 0:
                text = " ".join(texts)

                try:
                    from tokenizers import Tokenizer  # type: ignore

                    if self._cohere_embed_english_v3_tokenizer is None: # type: ignore
                        current_dir = os.path.dirname(os.path.abspath(__file__))
                        tokenizer_path = os.path.join(current_dir, "data", "cohere_embed_english_v3.json")
                        self._cohere_embed_english_v3_tokenizer = Tokenizer.from_file(tokenizer_path) # type: ignore

                    if self._cohere_embed_english_v3_tokenizer is not None and isinstance(self._cohere_embed_english_v3_tokenizer, Tokenizer): # type: ignore
                        tokens: list = self._cohere_embed_english_v3_tokenizer.encode(text, add_special_tokens=False).tokens # type: ignore

                        if tokens and isinstance(tokens, list):
                            units["text"] = IngestUnits(input=len(tokens), output=0) # type: ignore

                except ImportError:
                    self._request._instrumentor._logger.warning("tokenizers module not found, caller must install the tokenizers module. Cannot record text tokens for Cohere embed english v3")
                    pass
                except Exception as e:
                    self._request._instrumentor._logger.warning(f"Error processing Cohere embed english v3 response: {e}")
                    pass

        if self._log_prompt_and_response:
            ingest["provider_response_json"] = data.decode('utf-8') # type: ignore

        guardrails = response.get("amazon-bedrock-trace", {}).get("guardrail", {}).get("input", {})
        self._request.process_guardrails(guardrails)

        self._request.process_stop_action(response.get("amazon-bedrock-guardrailAction", ""))

        xproxy_result = self._request._instrumentor._ingest_units(self._request)
        self._request.assign_xproxy_result(self._response, xproxy_result)

        return data # type: ignore

def wrap_invoke(instrumentor: _PayiInstrumentor, wrapped: Any, instance: Any) -> Any:
    @wraps(wrapped)
    def invoke_wrapper(*args: Any, **kwargs: 'dict[str, Any]') -> Any:
        modelId:str = kwargs.get("modelId", "") # type: ignore

        return instrumentor.invoke_wrapper(
            _BedrockInvokeProviderRequest(instrumentor=instrumentor, model_id=modelId),
            _IsStreaming.false,
            wrapped,
            instance,
            args,
            kwargs,
        )   
    
    return invoke_wrapper

def wrap_invoke_stream(instrumentor: _PayiInstrumentor, wrapped: Any, instance: Any) -> Any:
    @wraps(wrapped)
    def invoke_wrapper(*args: Any, **kwargs: Any) -> Any:
        modelId: str = kwargs.get("modelId", "") # type: ignore

        instrumentor._logger.debug(f"bedrock invoke stream wrapper, modelId: {modelId}")
        return instrumentor.invoke_wrapper(
            _BedrockInvokeProviderRequest(instrumentor=instrumentor, model_id=modelId),
            _IsStreaming.true,
            wrapped,
            instance,
            args,
            kwargs,
        )

    return invoke_wrapper

def wrap_converse(instrumentor: _PayiInstrumentor, wrapped: Any, instance: Any) -> Any:
    @wraps(wrapped)
    def invoke_wrapper(*args: Any, **kwargs: 'dict[str, Any]') -> Any:
        modelId:str = kwargs.get("modelId", "") # type: ignore

        instrumentor._logger.debug(f"bedrock converse wrapper, modelId: {modelId}")
        return instrumentor.invoke_wrapper(
            _BedrockConverseProviderRequest(instrumentor=instrumentor, instance=instance),
            _IsStreaming.false,
            wrapped,
            instance,
            args,
            kwargs,
        )
    
    return invoke_wrapper

def wrap_converse_stream(instrumentor: _PayiInstrumentor, wrapped: Any, instance: Any) -> Any:
    @wraps(wrapped)
    def invoke_wrapper(*args: Any, **kwargs: Any) -> Any:
        modelId: str = kwargs.get("modelId", "") # type: ignore

        instrumentor._logger.debug(f"bedrock converse stream wrapper, modelId: {modelId}")
        return instrumentor.invoke_wrapper(
            _BedrockConverseProviderRequest(instrumentor=instrumentor, instance=instance),
            _IsStreaming.true,
            wrapped,
            instance,
            args,
            kwargs,
        )

    return invoke_wrapper

class _BedrockProviderRequest(_ProviderRequest):

    def __init__(self, instrumentor: _PayiInstrumentor, instance: Any = None) -> None:
        super().__init__(
            instrumentor=instrumentor,
            category=PayiCategories.aws_bedrock,
            streaming_type=_StreamingType.iterator,
            module_name=BedrockInstrumentor._module_name,
            module_version=BedrockInstrumentor._module_version,
            is_aws_client=True,
            )

        try:
            self.provider_uri = instance._endpoint.host
        except Exception:
            pass

    @override
    def process_request(self, instance: Any, args: Sequence[Any], kwargs: Any) -> bool:
        modelId =  kwargs.get("modelId", "")
        self._ingest["resource"] = modelId

        if not self._price_as.resource and not self._price_as.category and modelId in BedrockInstrumentor._model_mapping:
            deployment = BedrockInstrumentor._model_mapping.get(modelId, {})
            self._price_as.category = deployment.get("price_as_category", "")
            self._price_as.resource = deployment.get("price_as_resource", "")
            self._price_as.resource_scope = deployment.get("resource_scope", None)

        if self._price_as.resource_scope:
            self._ingest["resource_scope"] = self._price_as.resource_scope
        
        # override defaults
        if self._price_as.category:
            self._ingest["category"] = self._price_as.category
        if self._price_as.resource:
            self._ingest["resource"] = self._price_as.resource

        return True

    def process_response_metadata(self, metadata: 'dict[str, Any]') -> None:
        request_id = metadata.get("RequestId", "")
        if request_id:
            self._ingest["provider_response_id"] = request_id

        response_headers = metadata.get("HTTPHeaders", {})
        if response_headers:
            self.add_response_headers(response_headers)

    @override
    def process_initial_stream_response(self, response: Any) -> None:
        super().process_initial_stream_response(response)
        self.process_response_metadata(response.get("ResponseMetadata", {}))

    @override
    def process_exception(self, exception: Exception, kwargs: Any, ) -> bool:
        try:
            if hasattr(exception, "response"):
                response: dict[str, Any] = getattr(exception, "response", {})
                status_code: int = response.get('ResponseMetadata', {}).get('HTTPStatusCode', 0)
                if status_code == 0:
                    return False

                self._ingest["http_status_code"] = status_code
                
                request_id = response.get('ResponseMetadata', {}).get('RequestId', "")
                if request_id:
                    self._ingest["provider_response_id"] = request_id

                error = response.get('Error', "")
                if error:
                    self._ingest["provider_response_json"] = json.dumps(error)

            return True

        except Exception as e:
            self._instrumentor._logger.debug(f"Error processing exception: {e}")
            return False

    def process_guardrails(self, guardrails: 'dict[str, Any]') -> None:
        units = self._ingest["units"]

        # while we iterate over the entire dict, only one guardrail is expected and supported
        for _, value in guardrails.items():
            # _ (key) is the guardrail id
            if not isinstance(value, dict):
                continue

            usage: dict[str, int] = value.get("invocationMetrics", {}).get("usage", {}) # type: ignore
            if not usage:
                continue

            topicPolicyUnits: int  = usage.get("topicPolicyUnits", 0) # type: ignore
            if topicPolicyUnits > 0:
                units["guardrail_topic"] = IngestUnits(input=topicPolicyUnits, output=0) # type: ignore

            contentPolicyUnits = usage.get("contentPolicyUnits", 0) # type: ignore
            if contentPolicyUnits > 0:
                units["guardrail_content"] = IngestUnits(input=contentPolicyUnits, output=0) # type: ignore

            wordPolicyUnits = usage.get("wordPolicyUnits", 0) # type: ignore    
            if wordPolicyUnits > 0:
                units["guardrail_word_free"] = IngestUnits(input=wordPolicyUnits, output=0) # type: ignore

            automatedReasoningPolicyUnits = usage.get("automatedReasoningPolicyUnits", 0) # type: ignore
            if automatedReasoningPolicyUnits > 0:
                units["guardrail_automated_reasoning"] = IngestUnits(input=automatedReasoningPolicyUnits, output=0) # type: ignore

            sensitiveInformationPolicyUnits = usage.get("sensitiveInformationPolicyUnits", 0) # type: ignore
            if sensitiveInformationPolicyUnits > 0:
                units["guardrail_sensitive_information"] = IngestUnits(input=sensitiveInformationPolicyUnits, output=0) # type: ignore

            sensitiveInformationPolicyFreeUnits = usage.get("sensitiveInformationPolicyFreeUnits", 0) # type: ignore
            if sensitiveInformationPolicyFreeUnits > 0:
                units["guardrail_sensitive_information_free"] = IngestUnits(input=sensitiveInformationPolicyFreeUnits, output=0) # type: ignore

            contextualGroundingPolicyUnits = usage.get("contextualGroundingPolicyUnits", 0) # type: ignore
            if contextualGroundingPolicyUnits > 0:
                units["guardrail_contextual_grounding"] = IngestUnits(input=contextualGroundingPolicyUnits, output=0) # type: ignore

            contentPolicyImageUnits = usage.get("contentPolicyImageUnits", 0) # type: ignore
            if contentPolicyImageUnits > 0:
                units["guardrail_content_image"] = IngestUnits(input=contentPolicyImageUnits, output=0) # type: ignore

class _BedrockInvokeProviderRequest(_BedrockProviderRequest):
    def __init__(self, instrumentor: _PayiInstrumentor, model_id: str):
        super().__init__(instrumentor=instrumentor)

        price_as_resource = BedrockInstrumentor._model_mapping.get(model_id, {}).get("price_as_resource", None)
        if price_as_resource:
            model_id = price_as_resource

        self._is_anthropic: bool = False
        self._is_nova: bool = False
        self._is_meta: bool = False
        self._is_amazon_titan_embed_text_v1: bool = False
        self._is_cohere_embed_english_v3: bool = False

        self._assign_model_state(model_id=model_id)

    def _assign_model_state(self, model_id: str) -> None:
        self._is_anthropic = 'anthropic' in model_id
        self._is_nova = 'nova' in model_id
        self._is_meta = 'meta' in model_id
        self._is_amazon_titan_embed_text_v1 = 'amazon.titan-embed-text-v1' == model_id
        self._is_cohere_embed_english_v3 = 'cohere.embed-english-v3' == model_id

    @override
    def process_request(self, instance: Any, args: Sequence[Any], kwargs: Any) -> bool:
        from .AnthropicInstrumentor import anthropic_has_image_and_get_texts

        super().process_request(instance, args, kwargs)
    
        # super().process_request will assign price_as mapping from global state, so evaluate afterwards
        if self._price_as.resource:
            self._assign_model_state(model_id=self._price_as.resource)

        guardrail_id = kwargs.get("guardrailIdentifier", "")
        if guardrail_id:
            self.add_internal_request_property(PayiPropertyNames.aws_bedrock_guardrail_id, guardrail_id)

        guardrail_version = kwargs.get("guardrailVersion", "")
        if guardrail_version:
            self.add_internal_request_property(PayiPropertyNames.aws_bedrock_guardrail_version, guardrail_version)

        if guardrail_id and guardrail_version and BedrockInstrumentor._guardrail_trace:
            trace = kwargs.get("trace", None)
            if not trace:
                kwargs["trace"] = "ENABLED"

        if self._is_anthropic:
            try:
                body = json.loads(kwargs.get("body", ""))
                messages = body.get("messages", {})
                if messages:
                    anthropic_has_image_and_get_texts(self, messages)
            except Exception as e:
                self._instrumentor._logger.debug(f"Bedrock invoke error processing request body: {e}")
        elif self._is_cohere_embed_english_v3:
            try:
                body = json.loads(kwargs.get("body", ""))
                input_type = body.get("input_type", "")
                if input_type == 'image':
                    images = body.get("images", [])
                    if (len(images) > 0):
                        # only supports one image according to docs
                        self._ingest["units"]["vision"] = IngestUnits(input=1, output=0)
            except Exception as e:
                self._instrumentor._logger.debug(f"Bedrock invoke error processing request body: {e}")
        return True

    @override
    def process_chunk(self, chunk: Any) -> _ChunkResult:
        chunk_dict = json.loads(chunk)

        guardrails = chunk_dict.get("amazon-bedrock-trace", {}).get("guardrail", {}).get("input", {})
        if guardrails:
            self.process_guardrails(guardrails)
    
        self.process_stop_action(chunk_dict.get("amazon-bedrock-guardrailAction", ""))

        if self._is_anthropic:
            from .AnthropicInstrumentor import anthropic_process_chunk
            return anthropic_process_chunk(self, chunk_dict, assign_id=False)
        
        if self._is_nova:
            bedrock_converse_process_streaming_for_function_call(self, chunk_dict)

        # meta and nova
        return self.process_invoke_other_provider_chunk(chunk_dict)

    def process_invoke_other_provider_chunk(self, chunk_dict: 'dict[str, Any]') -> _ChunkResult:
        ingest = False

        metrics = chunk_dict.get("amazon-bedrock-invocationMetrics", {})
        if metrics:
            input = metrics.get("inputTokenCount", 0)
            output = metrics.get("outputTokenCount", 0)
            self._ingest["units"]["text"] = IngestUnits(input=input, output=output)

            text_cache_read = metrics.get("cacheReadInputTokenCount", None)
            if text_cache_read:
                self._ingest["units"]["text_cache_read"] = text_cache_read

            text_cache_write = metrics.get("cacheWriteInputTokenCount", None)
            if text_cache_write:
                self._ingest["units"]["text_cache_write"] = text_cache_write

            ingest = True

        return _ChunkResult(send_chunk_to_caller=True, ingest=ingest)    

    @override
    def process_synchronous_response(
        self,
        response: Any,
        kwargs: Any) -> Any:

        self.process_response_metadata(response.get("ResponseMetadata", {}))

        response["body"] = InvokeResponseWrapper(
            response=response,
            body=response["body"],
            request=self)

        return response

    def process_stop_action(self, action: str) -> None:
        # record both as a semantic failure and guardrail action so it is discoverable through both properties
        if action == "INTERVENED":
            self.add_internal_request_property(PayiPropertyNames.failure, action)
            self.add_internal_request_property(PayiPropertyNames.failure_description, GUARDRAIL_SEMANTIC_FAILURE_DESCRIPTION)
            self.add_internal_request_property(PayiPropertyNames.aws_bedrock_guardrail_action, action)

    @override
    def remove_prompt_inline_data(self, prompt: 'dict[str, Any]') -> bool:# noqa: ARG002
        if not self._is_anthropic:
            return False

        from .AnthropicInstrumentor import anthropic_remove_inline_data
        body = prompt.get("body", "")
        if not body:
            return False
        
        body_json = json.loads(body)
        
        if anthropic_remove_inline_data(body_json):
            prompt["body"] = json.dumps(body_json)
            return True

        return False

class _BedrockConverseProviderRequest(_BedrockProviderRequest):
    @override
    def process_request(self, instance: Any, args: Sequence[Any], kwargs: Any) -> bool:
        super().process_request(instance, args, kwargs)

        guardrail_config = kwargs.get("guardrailConfig", {})
        if guardrail_config:
            guardrailIdentifier = guardrail_config.get("guardrailIdentifier", "")
            if guardrailIdentifier:
                self.add_internal_request_property(PayiPropertyNames.aws_bedrock_guardrail_id, guardrailIdentifier)

            guardrailVersion = guardrail_config.get("guardrailVersion", "")
            if guardrailVersion:
                self.add_internal_request_property(PayiPropertyNames.aws_bedrock_guardrail_version, guardrailVersion)

            if guardrailIdentifier and guardrailVersion and BedrockInstrumentor._guardrail_trace:
                trace = guardrail_config.get("trace", None)
                if not trace:
                    guardrail_config["trace"] = "enabled"

        return True

    @override
    def process_synchronous_response(
        self,
        response: 'dict[str, Any]',
        kwargs: Any) -> Any:

        usage = response.get("usage", {})
        input = usage.get("inputTokens", 0)
        output = usage.get("outputTokens", 0)

        units: dict[str, IngestUnits] = self._ingest["units"]
        units["text"] = IngestUnits(input=input, output=output)

        self.process_response_metadata(response.get("ResponseMetadata", {}))

        if self._log_prompt_and_response:
            response_without_metadata = response.copy()
            response_without_metadata.pop("ResponseMetadata", None)
            self._ingest["provider_response_json"] = json.dumps(response_without_metadata)

        bedrock_converse_process_synchronous_function_call(self, response)

        guardrails = response.get("trace", {}).get("guardrail", {}).get("inputAssessment", {})
        if guardrails:
            self.process_guardrails(guardrails)

        self.process_stop_reason(response.get("stopReason", ""))

        return None

    @override
    def process_chunk(self, chunk: 'dict[str, Any]') -> _ChunkResult:
        ingest = False
        metadata = chunk.get("metadata", {})

        if metadata:
            usage = metadata.get('usage', {})
            input = usage.get("inputTokens", 0)
            output = usage.get("outputTokens", 0)
            self._ingest["units"]["text"] = IngestUnits(input=input, output=output)

            guardrail = metadata.get("trace", {}).get("guardrail", {}).get("inputAssessment", {})
            if guardrail:
                self.process_guardrails(guardrail)

            ingest = True

        self.process_stop_reason(chunk.get("messageStop", {}).get("stopReason", ""))

        bedrock_converse_process_streaming_for_function_call(self, chunk)

        return _ChunkResult(send_chunk_to_caller=True, ingest=ingest)

    def process_stop_reason(self, reason: str) -> None:
        if reason == "guardrail_intervened":
            # record both as a semantic failure and guardrail action so it is discoverable through both properties
            self.add_internal_request_property(PayiPropertyNames.failure, reason)
            self.add_internal_request_property(PayiPropertyNames.failure_description, GUARDRAIL_SEMANTIC_FAILURE_DESCRIPTION)
            self.add_internal_request_property(PayiPropertyNames.aws_bedrock_guardrail_action, reason)

def bedrock_converse_process_streaming_for_function_call(request: _ProviderRequest, chunk: 'dict[str, Any]') -> None:  
    contentBlockStart = chunk.get("contentBlockStart", {})
    tool_use = contentBlockStart.get("start", {}).get("toolUse", {})
    if tool_use:
        index = contentBlockStart.get("contentBlockIndex", None)
        name = tool_use.get("name", "")

        if name and index is not None:
            request.add_streaming_function_call(index=index, name=name, arguments=None)
        
        return

    contentBlockDelta = chunk.get("contentBlockDelta", {})
    tool_use = contentBlockDelta.get("delta", {}).get("toolUse", {})
    if tool_use:
        index = contentBlockDelta.get("contentBlockIndex", None)
        input = tool_use.get("input", "")

        if input and index is not None:
            request.add_streaming_function_call(index=index, name=None, arguments=input)

        return

def bedrock_converse_process_synchronous_function_call(request: _ProviderRequest, response: 'dict[str, Any]') -> None:
    content = response.get("output", {}).get("message", {}).get("content", [])
    if content:
        for item in content:
            tool_use = item.get("toolUse", {})
            if tool_use:
                name = tool_use.get("name", "")
                input = tool_use.get("input", {})
                arguments: Optional[str] = None

                if input and isinstance(input, dict):
                    arguments = json.dumps(input)
                
                if name:
                    request.add_synchronous_function_call(name=name, arguments=arguments)




# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .result import (
    ResultResource,
    AsyncResultResource,
    ResultResourceWithRawResponse,
    AsyncResultResourceWithRawResponse,
    ResultResourceWithStreamingResponse,
    AsyncResultResourceWithStreamingResponse,
)
from .properties import (
    PropertiesResource,
    AsyncPropertiesResource,
    PropertiesResourceWithRawResponse,
    AsyncPropertiesResourceWithRawResponse,
    PropertiesResourceWithStreamingResponse,
    AsyncPropertiesResourceWithStreamingResponse,
)
from .response_id import (
    ResponseIDResource,
    AsyncResponseIDResource,
    ResponseIDResourceWithRawResponse,
    AsyncResponseIDResourceWithRawResponse,
    ResponseIDResourceWithStreamingResponse,
    AsyncResponseIDResourceWithStreamingResponse,
)

__all__ = [
    "ResultResource",
    "AsyncResultResource",
    "ResultResourceWithRawResponse",
    "AsyncResultResourceWithRawResponse",
    "ResultResourceWithStreamingResponse",
    "AsyncResultResourceWithStreamingResponse",
    "PropertiesResource",
    "AsyncPropertiesResource",
    "PropertiesResourceWithRawResponse",
    "AsyncPropertiesResourceWithRawResponse",
    "PropertiesResourceWithStreamingResponse",
    "AsyncPropertiesResourceWithStreamingResponse",
    "ResponseIDResource",
    "AsyncResponseIDResource",
    "ResponseIDResourceWithRawResponse",
    "AsyncResponseIDResourceWithRawResponse",
    "ResponseIDResourceWithStreamingResponse",
    "AsyncResponseIDResourceWithStreamingResponse",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List
from typing_extensions import TypeAlias

from ..limit_response import LimitResponse

__all__ = ["PropertyUpdateResponse"]

PropertyUpdateResponse: TypeAlias = List[LimitResponse]



"""Ingest retry manager for the Pay-i instrumentation layer.

Handles inline retry with exponential backoff and a background retry queue
for failed ingest calls.  Extracted from ``instrument.py`` so the
``_PayiInstrumentor`` class stays focused on instrumentation orchestration.
"""
from __future__ import annotations

import copy
import time
import atexit
import random
import asyncio
import logging
import threading
from typing import TYPE_CHECKING, Any, Union, Callable, Optional, Awaitable
from collections import deque

import httpx
import httpcore

from payi import APIStatusError, APIConnectionError
from payi.types import IngestUnitsParams
from payi.types.ingest_response import IngestResponse

if TYPE_CHECKING:
    from .instrument import PayiInstrumentIngestRetryConfig

# ---------------------------------------------------------------------------
# Retry defaults
# ---------------------------------------------------------------------------

# Inline-retry defaults for ingest calls at the instrumentation layer.
# These retries are on top of the SDK's built-in retries (DEFAULT_MAX_RETRIES=2).
# When httpx connection pools contain stale connections (common behind Azure App Gateway
# or other reverse proxies), the SDK's retries may all hit stale connections from the
# same pool. The instrumentation-layer retry gives the pool time to evict dead connections,
# so subsequent attempts are more likely to get a fresh connection.
# Callers can override via PayiInstrumentConfig keys ingest_max_retries / ingest_retry_initial_delay.
_INGEST_MAX_RETRIES_DEFAULT = 0  # 2
_INGEST_RETRY_INITIAL_DELAY_DEFAULT = 0.5  # seconds, doubles on each retry

# Transport-level exceptions that indicate a stale or broken pooled connection.
# Only these warrant an instrumentation-layer retry — other APIConnectionError causes
# (e.g. DNS resolution failure, TLS certificate errors) should not be retried here.
_RETRYABLE_CONNECTION_EXCEPTIONS = (
    httpx.TimeoutException,
    httpx.ReadError,
    httpx.WriteError,
    httpx.RemoteProtocolError,
    httpx.ConnectError,

    httpcore.TimeoutException,
    httpcore.ReadError,
    httpcore.WriteError,
    httpcore.RemoteProtocolError,
    httpcore.ConnectError,

    ConnectionResetError,
)

# Retry-queue defaults for background re-delivery of failed ingest calls.
_RETRY_QUEUE_MAX_SIZE_DEFAULT = 0          # 0 = unlimited (queue.Queue convention)
_RETRY_QUEUE_WORKER_INTERVAL_DEFAULT = 5.0  # seconds between drain cycles
_RETRY_QUEUE_WORKER_JITTER = 0.5           # +/- seconds applied to each sleep
_RETRY_QUEUE_DRAIN_TIMEOUT = 10.0          # seconds to wait for final drain on shutdown
_RETRY_QUEUE_BATCH_WINDOW = 2.0            # seconds to keep sending within a single drain cycle


# ---------------------------------------------------------------------------
# Helper used by both the manager and callers in instrument.py
# ---------------------------------------------------------------------------

def _qualified_exception_name(exc: BaseException | None) -> str:
    if exc is None:
        return "None"
    exc_type = type(exc)
    return f"{exc_type.__module__}.{exc_type.__qualname__}"


def is_retryable_connection_error(e: Union[APIConnectionError, APIStatusError]) -> bool:
    """Check whether an error is retryable at the instrumentation layer.

    Returns True for transport-level APIConnectionErrors (stale pooled
    connections) and for 504 Gateway Timeout APIStatusErrors (App Gateway
    upstream timeout — functionally equivalent to a stale connection).
    """
    if isinstance(e, APIStatusError):
        return e.status_code == 504
    cause = e.__cause__
    return isinstance(cause, _RETRYABLE_CONNECTION_EXCEPTIONS)


# ---------------------------------------------------------------------------
# Queue item
# ---------------------------------------------------------------------------

class _RetryQueueItem:
    """An ingest call that exhausted immediate retries and is awaiting background retry."""
    __slots__ = ('ingest_units', 'extra_headers', 'retry_count', 'query')

    def __init__(self, ingest_units: IngestUnitsParams, extra_headers: 'dict[str, str]', query: Optional[dict[str, str]]) -> None:
        self.ingest_units = ingest_units
        self.extra_headers = extra_headers
        self.retry_count = 0
        self.query = query


# ---------------------------------------------------------------------------
# Callable type aliases for dependency injection
# ---------------------------------------------------------------------------

SyncIngestFn = Callable[..., IngestResponse]
AsyncIngestFn = Callable[..., Awaitable[IngestResponse]]
OnSuccessFn = Callable[[IngestResponse], Any]
OnConnectionErrorFn = Callable[[APIConnectionError, IngestUnitsParams], Any]
OnApiStatusErrorFn = Callable[[APIStatusError], Any]


# ---------------------------------------------------------------------------
# IngestRetryManager
# ---------------------------------------------------------------------------

class IngestRetryManager:
    """Manages inline retry and background retry queue for failed ingest calls.

    All Pay-i / AsyncPayi client dependencies are injected as callables so this
    class has no direct reference to the client objects.
    """

    def __init__(
        self,
        *,
        sync_ingest_fn: Optional[SyncIngestFn] = None,
        async_ingest_fn: Optional[AsyncIngestFn] = None,
        on_success: OnSuccessFn,
        on_connection_error: OnConnectionErrorFn,
        on_api_status_error: OnApiStatusErrorFn,
        config: PayiInstrumentIngestRetryConfig,
        logger: Optional[logging.Logger] = None,
    ):
        self._logger: logging.Logger = logger if logger else logging.getLogger("payi.instrument")
        self._retry_logger: logging.Logger = self._logger.getChild("retry")

        self._sync_ingest_fn = sync_ingest_fn
        self._async_ingest_fn = async_ingest_fn
        self._on_success = on_success
        self._on_connection_error = on_connection_error
        self._on_api_status_error = on_api_status_error

        # --- Inline retry config ---
        ingest_max_retries = config.get("max_inline_retries", None)
        ingest_max_retries = ingest_max_retries if ingest_max_retries is not None else _INGEST_MAX_RETRIES_DEFAULT
        self._ingest_max_retries: int = ingest_max_retries
        if self._ingest_max_retries < 0:
            raise ValueError("ingest_max_retries must be a non-negative integer")

        ingest_initial_delay = config.get("inline_retry_initial_delay", None)
        ingest_initial_delay = ingest_initial_delay if ingest_initial_delay is not None else _INGEST_RETRY_INITIAL_DELAY_DEFAULT
        self._ingest_retry_initial_delay: float = ingest_initial_delay
        if self._ingest_retry_initial_delay < 0:
            raise ValueError("ingest_retry_initial_delay must be a non-negative number")

        # --- Retry queue for background re-delivery of failed ingest calls ---
        queue_enabled = config.get("queue_enabled", None)
        self._ingest_retry_queue_enabled: bool = queue_enabled if queue_enabled is not None else True

        if self._ingest_retry_queue_enabled:
            queue_max_size = config.get("queue_max_size", None)
            self._retry_queue_max_size: int = queue_max_size if queue_max_size is not None else _RETRY_QUEUE_MAX_SIZE_DEFAULT
            if self._retry_queue_max_size < 0:
                raise ValueError("queue_max_size must be a non-negative integer")

            queue_interval = config.get("queue_interval", None)
            self._retry_queue_interval: float = queue_interval if queue_interval is not None else _RETRY_QUEUE_WORKER_INTERVAL_DEFAULT
            if self._retry_queue_interval <= 0:
                raise ValueError("queue_interval must be a positive number")

            self._retry_queue: deque[_RetryQueueItem] = deque()
            self._retry_queue_lock: threading.Lock = threading.Lock()
            self._retry_queue_worker_thread: Optional[threading.Thread] = None
            self._retry_queue_shutdown: threading.Event = threading.Event()
            self._retry_queue_started_lock: threading.Lock = threading.Lock()

    # ------------------------------------------------------------------
    # Inline retry (sync / async)
    # ------------------------------------------------------------------

    def ingest_with_inline_retry(
        self,
        ingest_units: IngestUnitsParams,
        extra_headers: 'dict[str, str]',
        query: Optional[dict[str, str]],
        no_retry: bool = False,
    ) -> IngestResponse:
        """Call sync ingest.units() with retry on transport-level connection errors.

        When the httpx connection pool contains stale connections (e.g. closed by an Azure App
        Gateway idle timeout), the SDK's built-in retries may all draw from the same poisoned
        pool.  This outer retry gives the pool time to evict dead connections between attempts.

        Retries on known transport errors (ReadError, WriteError, RemoteProtocolError,
        ConnectError, ConnectionResetError) and on 504 Gateway Timeout.  Other errors
        are raised immediately.
        """
        max_retries = self._ingest_max_retries if not no_retry else 0
        last_error: Optional[Union[APIConnectionError, APIStatusError]] = None

        for attempt in range(1 + max_retries):
            try:
                return self._sync_ingest_fn(**ingest_units, extra_headers=extra_headers, extra_query=query)  # type: ignore[misc]
            except (APIConnectionError, APIStatusError) as e:
                if not is_retryable_connection_error(e):
                    raise
                last_error = e
                if attempt < max_retries:
                    delay = self._ingest_retry_initial_delay * (2 ** attempt)
                    self._retry_logger.warning(
                        f"Pay-i ingest retryable error (attempt {attempt + 1}/{1 + max_retries}), "
                        f"retrying in {delay:.1f}s: {_qualified_exception_name(e.__cause__) if e.__cause__ else e}"
                    )
                    time.sleep(delay)
                    continue
                raise

        raise last_error  # type: ignore[misc]

    async def aingest_with_inline_retry(
        self,
        ingest_units: IngestUnitsParams,
        extra_headers: 'dict[str, str]',
        query: Optional[dict[str, str]],
        no_retry: bool = False,
    ) -> IngestResponse:
        """Async variant of ingest_with_inline_retry."""
        max_retries = self._ingest_max_retries if not no_retry else 0
        last_error: Optional[Union[APIConnectionError, APIStatusError]] = None

        for attempt in range(1 + max_retries):
            try:
                return await self._async_ingest_fn(**ingest_units, extra_headers=extra_headers, extra_query=query)  # type: ignore[misc]
            except (APIConnectionError, APIStatusError) as e:
                if not is_retryable_connection_error(e):
                    raise
                last_error = e
                if attempt < max_retries:
                    delay = self._ingest_retry_initial_delay * (2 ** attempt)
                    self._retry_logger.warning(
                        f"Pay-i ingest retryable error (attempt {attempt + 1}/{1 + max_retries}), "
                        f"retrying in {delay:.1f}s: {_qualified_exception_name(e.__cause__) if e.__cause__ else e}"
                    )
                    await asyncio.sleep(delay)
                    continue
                raise

        raise last_error  # type: ignore[misc]

    # ------------------------------------------------------------------
    # Background retry queue
    # ------------------------------------------------------------------

    def enqueue_failed_ingest(
        self,
        ingest_units: IngestUnitsParams,
        extra_headers: 'dict[str, str]',
        query: Optional[dict[str, str]],
        api_ex: Union[APIConnectionError, APIStatusError],
    ) -> bool:
        """Enqueue a failed ingest call for background retry.

        Returns True if the item was enqueued, False if the queue is full.
        Deep-copies ingest_units to prevent mutation by the caller after enqueue.
        """
        if not self._ingest_retry_queue_enabled:
            return False

        item = _RetryQueueItem(
            ingest_units=copy.deepcopy(ingest_units),
            extra_headers=extra_headers.copy(),
            query=query.copy() if query else None,
        )

        with self._retry_queue_lock:
            if self._retry_queue_max_size > 0 and len(self._retry_queue) >= self._retry_queue_max_size:
                self._retry_logger.warning(
                    "Pay-i retry queue is full (max_size=%d), dropping ingest request",
                    self._retry_queue_max_size,
                )
                return False
            self._retry_queue.append(item)

        self._retry_logger.error(f"Delayed retry of Pay-i ingest exception {api_ex}, cause {_qualified_exception_name(api_ex.__cause__)}, request {ingest_units}, query {query}")

        self._retry_logger.debug(
            "Pay-i enqueued failed ingest for background retry (queue depth ~%d)",
            len(self._retry_queue),
        )

        # Lazily start the worker thread on first enqueue
        self._ensure_retry_queue_worker()
        return True

    def _ensure_retry_queue_worker(self) -> None:
        """Lazily start the background retry-queue worker thread (once)."""
        with self._retry_queue_started_lock:
            if self._retry_queue_worker_thread is not None:
                return

            thread = threading.Thread(
                target=self._retry_queue_worker_loop,
                name="payi-retry-queue-worker",
                daemon=True,
            )
            thread.start()
            self._retry_queue_worker_thread = thread
            self._retry_logger.debug("Pay-i retry queue worker thread started")

            atexit.register(self.shutdown)

    def _retry_queue_worker_loop(self) -> None:
        """Background thread that periodically drains and retries failed ingest calls."""
        self._retry_logger.debug("Pay-i retry queue worker loop started")

        while not self._retry_queue_shutdown.is_set():
            jitter = random.uniform(-_RETRY_QUEUE_WORKER_JITTER, _RETRY_QUEUE_WORKER_JITTER)
            sleep_time = max(0.1, self._retry_queue_interval + jitter)

            # Use Event.wait() so shutdown can interrupt the sleep
            if self._retry_queue_shutdown.wait(timeout=sleep_time):
                break  # Shutdown signaled during sleep

            self._drain_retry_queue()

        self._retry_logger.debug("Pay-i retry queue worker loop exited")

    def _drain_retry_queue(self) -> None:
        """Peek at front, try to send, pop only on success or non-retryable error.

        Sends as many items as possible within a time window.  On retryable
        failure the item stays at the front with its retry_count bumped and the
        batch stops.
        """
        deadline = time.monotonic() + _RETRY_QUEUE_BATCH_WINDOW
        sent = 0

        while time.monotonic() < deadline:
            with self._retry_queue_lock:
                if not self._retry_queue:
                    break
                item = self._retry_queue[0]  # peek

            if not self._retry_single_queued_item(item):
                break  # failed — item stays at front, next cycle retries

            # Success — remove the item we just sent
            with self._retry_queue_lock:
                if self._retry_queue and self._retry_queue[0] is item:
                    self._retry_queue.popleft()
            sent += 1

        if sent:
            self._retry_logger.debug("Pay-i retry queue batch sent %d items", sent)

    def _retry_single_queued_item(self, item: _RetryQueueItem) -> bool:
        """Attempt to re-send a single queued ingest call.

        Returns True on success, False on failure.  On retryable failure the
        item stays in the queue with its retry_count bumped.  On non-retryable
        error the item is removed from the queue and dropped.
        """
        try:
            if self._sync_ingest_fn:
                response = self.ingest_with_inline_retry(item.ingest_units, item.extra_headers, query=item.query, no_retry=True)
            else:
                response = self._run_async_ingest_from_thread(item.ingest_units, item.extra_headers, query=item.query)

            self._retry_logger.debug(
                "Pay-i retry queue: successfully re-sent ingest request (retry_count=%d)",
                item.retry_count,
            )
            if response:
                self._on_success(response)
            return True

        except (APIConnectionError, APIStatusError) as e:
            if is_retryable_connection_error(e):
                item.retry_count += 1
                self._retry_logger.debug(
                    "Pay-i retry queue: kept at front (retry_count=%d)",
                    item.retry_count,
                )
            else:
                # Non-retryable — remove and drop
                with self._retry_queue_lock:
                    if self._retry_queue and self._retry_queue[0] is item:
                        self._retry_queue.popleft()
                if isinstance(e, APIConnectionError):
                    self._on_connection_error(e, item.ingest_units)
                else:
                    self._on_api_status_error(e)
            return False

        except Exception as e:
            with self._retry_queue_lock:
                if self._retry_queue and self._retry_queue[0] is item:
                    self._retry_queue.popleft()
            self._retry_logger.error(
                "Pay-i retry queue: unexpected error retrying ingest (retry_count=%d): %s",
                item.retry_count, e,
            )
            return False

    def _run_async_ingest_from_thread(
        self,
        ingest_units: IngestUnitsParams,
        extra_headers: 'dict[str, str]',
        query: Optional[dict[str, str]],
    ) -> IngestResponse:
        """Run async ingest-with-retry from the background worker thread."""
        loop = asyncio.new_event_loop()
        try:
            return loop.run_until_complete(
                self.aingest_with_inline_retry(ingest_units, extra_headers, query=query, no_retry=True)
            )
        finally:
            loop.close()

    # ------------------------------------------------------------------
    # Shutdown
    # ------------------------------------------------------------------

    def shutdown(self) -> None:
        """Signal the worker thread to stop, then try to send remaining items once."""
        if not self._ingest_retry_queue_enabled:
            return

        if self._retry_queue_worker_thread is None:
            return

        self._logger.debug("Pay-i retry queue: initiating shutdown")
        self._retry_queue_shutdown.set()

        self._retry_queue_worker_thread.join(timeout=_RETRY_QUEUE_DRAIN_TIMEOUT)

        if self._retry_queue_worker_thread.is_alive():
            self._logger.warning("Pay-i retry queue worker did not exit within %.1f seconds", _RETRY_QUEUE_DRAIN_TIMEOUT)

        # Final best-effort drain: send each item once without retry,
        # stop on the first APIConnectionError.
        with self._retry_queue_lock:
            items = list(self._retry_queue)
            self._retry_queue.clear()

        sent = 0
        remaining = len(items)
        for item in items:
            try:
                if self._sync_ingest_fn:
                    self._sync_ingest_fn(**item.ingest_units, extra_headers=item.extra_headers, query=item.query)
                else:
                    self._run_async_ingest_from_thread(item.ingest_units, item.extra_headers, query=item.query)
                sent += 1
            except APIConnectionError:
                self._logger.debug("Pay-i retry queue shutdown: connection error, stopping final drain")
                break
            except Exception as e:
                self._logger.debug("Pay-i retry queue shutdown: error sending item: %s", e)

        if remaining > 0:
            self._logger.debug("Pay-i retry queue shutdown: sent %d/%d remaining items", sent, remaining)



from __future__ import annotations

import inspect
from abc import abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Any, Union, Optional, Sequence
from dataclasses import dataclass

import httpx

from payi.types import IngestUnitsParams
from payi.lib.helpers import PayiPropertyNames
from payi.types.ingest_units_params import ProviderResponseFunctionCall
from payi.types.shared.xproxy_error import XproxyError
from payi.types.shared.xproxy_result import XproxyResult
from payi.types.shared_params.ingest_units import IngestUnits
from payi.types.pay_i_common_models_api_router_header_info_param import PayICommonModelsAPIRouterHeaderInfoParam

from .helpers import _set_attr_safe
from .Stopwatch import Stopwatch

if TYPE_CHECKING:
    from .instrument import _PayiInstrumentor

class _StreamingType(Enum):
    generator = 0
    iterator = 1
    stream_manager = 2

@dataclass
class _ChunkResult:
    send_chunk_to_caller: bool
    ingest: bool = False

@dataclass
class PriceAs:
    category: Optional[str]
    resource: Optional[str]
    resource_scope: Optional[str]

class _ProviderRequest:
    excluded_headers = {
        "transfer-encoding",
    }

    _instrumented_response_headers_attr = "_instrumented_response_headers"
    _xproxy_result_attr = "xproxy_result"

    def __init__(
            self,
            instrumentor: _PayiInstrumentor,
            category: str,
            streaming_type: _StreamingType,
            module_name: str,
            module_version: str,
            is_aws_client: Optional[bool] = None,
            is_google_vertex_or_genai_client: Optional[bool] = None,
            ) -> None:
        self._instrumentor: _PayiInstrumentor = instrumentor
        self._module_name: str = module_name
        self._module_version: str = module_version  
        self._estimated_prompt_tokens: Optional[int] = None
        self._category: str = category
        self._ingest: IngestUnitsParams = { "category": category, "units": {} } # type: ignore
        self._streaming_type: '_StreamingType' = streaming_type
        self._is_aws_client: Optional[bool] = is_aws_client
        self._is_google_vertex_or_genai_client: Optional[bool] = is_google_vertex_or_genai_client
        self._function_call_builder: Optional[dict[int, ProviderResponseFunctionCall]] = None
        self._building_function_response: bool = False
        self._function_calls: Optional[list[ProviderResponseFunctionCall]] = None
        self._is_large_context: bool = False
        self._internal_request_properties: dict[str, Optional[str]] = {}
        self._price_as: PriceAs = PriceAs(category=None, resource=None, resource_scope=None)
        self._log_prompt_and_response: bool = instrumentor._log_prompt_and_response
        self._stopwatch: Stopwatch = Stopwatch()

    def process_chunk(self, _chunk: Any) -> _ChunkResult:
        return _ChunkResult(send_chunk_to_caller=True)

    def process_synchronous_response(self, response: Any, kwargs: Any) -> Optional[object]:  # noqa: ARG002
        return None
    
    @abstractmethod
    def process_request(self, instance: Any, args: Sequence[Any], kwargs: Any) -> bool:
        ...
    
    def process_request_prompt(self, prompt: 'dict[str, Any]', args: Sequence[Any], kwargs: 'dict[str, Any]') -> None:
        ...
    
    def process_initial_stream_response(self, response: Any) -> None:
        self.add_instrumented_response_headers(response)

    def remove_prompt_inline_data(self, prompt: 'dict[str, Any]') -> bool:# noqa: ARG002
        return False

    def remove_responses_inline_data(self, responses: 'list[dict[str, Any]]') -> bool:# noqa: ARG002
        return False
    
    def get_host(self, url: Union[str, httpx.URL, Any]) -> Optional[str]:
        try:
            if isinstance(url, str):
                url = httpx.URL(url)
            elif not isinstance(url, httpx.URL):
                return None
            return url.host
        except Exception:
            return None

    def category_from_host(self, host: str) -> str:
        return self._instrumentor._host_mappings.get(host, self._category)

    @property
    def provider_uri(self) -> Optional[str]:
        return self._ingest.get("provider_uri") if "provider_uri" in self._ingest else None
    
    @provider_uri.setter
    def provider_uri(self, value: str) -> None:
        self._ingest["provider_uri"] = value
        host = self.get_host(value)
        if host:
            self._category = self.category_from_host(host)
            self._ingest["category"] = self._category

    @property
    def stopwatch(self) -> Stopwatch:
        return self._stopwatch
    
    @property
    def is_aws_client(self) -> bool:
        return self._is_aws_client if self._is_aws_client is not None else False

    @property
    def is_google_vertex_or_genai_client(self) -> bool:
        return self._is_google_vertex_or_genai_client if self._is_google_vertex_or_genai_client is not None else False

    def process_exception(self, exception: Exception, kwargs: Any, ) -> bool: # noqa: ARG002
        self.exception_to_semantic_failure(exception)
        return True
    
    @property
    def supports_extra_headers(self) -> bool:
        return not self.is_aws_client and not self.is_google_vertex_or_genai_client
    
    @property
    def streaming_type(self) -> '_StreamingType':
        return self._streaming_type

    def add_internal_request_property(self, key: str, value: str) -> None:
        self._internal_request_properties[key] = value

    def exception_to_semantic_failure(self, e: Exception) -> None:
        exception_str = f"{type(e).__name__}"
    
        fields: list[str] = []
    
        for attr in dir(e):
            if not attr.startswith("__"):
                try:
                    value = getattr(e, attr)
                    if value and not inspect.ismethod(value) and not inspect.isfunction(value) and not callable(value):
                        fields.append(f"{attr}={value}")
                except Exception as _ex:
                    pass
 
        self.add_internal_request_property(PayiPropertyNames.failure, exception_str)
        if fields:
            failure_description = ",".join(fields)
            self.add_internal_request_property(PayiPropertyNames.failure_description, failure_description)

        if "http_status_code" not in self._ingest:
            # use a non existent http status code so when presented to the user, the origin is clear
            self._ingest["http_status_code"] = 299

    def add_streaming_function_call(self, index: int, name: Optional[str], arguments: Optional[str]) -> None:
        if not self._function_call_builder:
            self._function_call_builder = {}

        if not index in self._function_call_builder:
            self._function_call_builder[index] = ProviderResponseFunctionCall(name=name or "", arguments=arguments or "")
        else:
            function = self._function_call_builder[index]
            if name:
                function["name"] = function["name"] + name
            if arguments:
                function["arguments"] = (function.get("arguments", "") or "") + arguments

    def add_synchronous_function_call(self, name: str, arguments: Optional[str]) -> None:
        if not self._function_calls:
            self._function_calls = []
            self._ingest["provider_response_function_calls"] = self._function_calls
        self._function_calls.append(ProviderResponseFunctionCall(name=name, arguments=arguments))
    
    def add_instrumented_response_headers(self, response: Any) -> None:
        response_headers  = getattr(response, _ProviderRequest._instrumented_response_headers_attr, {})
        if response_headers:
            self.add_response_headers(response_headers)

    def add_response_headers(self, response_headers: 'dict[str, Any]') -> None:
        self._ingest["provider_response_headers"] = [
            PayICommonModelsAPIRouterHeaderInfoParam(name=k, value=v) 
            for k, v in response_headers.items() 
            if (k_lower := k.lower()) not in _ProviderRequest.excluded_headers and not k_lower.startswith("content-")
        ]

    def find_response_header_value(self, header_name: str) -> Optional[str]:
        response_headers = self._ingest.get("provider_response_headers", None)
        if response_headers:
            header_name = header_name.lower()
            for header in response_headers:
                if header.get("name", "").lower() == header_name:
                    return header.get("value", None)
        return None

    def merge_internal_request_properties(self) -> None:
        if not self._internal_request_properties:
            return
        
        properties = self._ingest.get("properties") or {}
        self._ingest["properties"] = properties
        for key, value in self._internal_request_properties.items():
            if key not in properties:
                properties[key] = value
                
    def update_for_vision(self, input: int) -> int:
        if self._estimated_prompt_tokens:
            vision = input - self._estimated_prompt_tokens
            if (vision > 0):
                key = "vision_large_context" if self._is_large_context else "vision"
                self._ingest["units"][key] = IngestUnits(input=vision, output=0)
                input = self._estimated_prompt_tokens
        
        return input

    @staticmethod
    def assign_xproxy_result(o: Any, xproxy_result: XproxyResult |  XproxyError| None) -> None:
        if xproxy_result:
            _set_attr_safe(o, _ProviderRequest._xproxy_result_attr, xproxy_result)

    @staticmethod
    def process_response_wrapper(wrapped: Any, _instance: Any, args: Any, kwargs: Any) -> Any:
        httpResponse = kwargs.get("response", None)

        r =  wrapped(*args, **kwargs)

        if httpResponse:
            headers = getattr(httpResponse, "headers", None)
            _set_attr_safe(r, _ProviderRequest._instrumented_response_headers_attr, dict(headers) if headers else {})

        return r

    @staticmethod
    async def aprocess_response_wrapper(wrapped: Any, _instance: Any, args: Any, kwargs: Any) -> Any:
        httpResponse = kwargs.get("response", None)

        r = await wrapped(*args, **kwargs)

        if httpResponse:
            headers = getattr(httpResponse, "headers", None)
            _set_attr_safe(r, _ProviderRequest._instrumented_response_headers_attr, dict(headers) if headers else {})

        return r

